import torch

def get_extrap_pos(xqextrap, extrap, xmin=0.0, xmax=1.0):
    # xqextrap: (nrq,)
    xqnorm = (xqextrap - xmin) / (xmax - xmin)
    if extrap == "periodic":
        xqinside = xqnorm % 1.0
    elif extrap == "mirror":
        xqnorm = xqnorm.abs()
        xqnorm_ceil = xqnorm.long() + 1
        xqhalf = torch.div(xqnorm_ceil, 2, rounding_mode="trunc")
        xqinside = (2 * xqhalf - xqnorm) * (1 - (xqnorm_ceil % 2.0) * 2)
    elif extrap == "bound":
        xqinside = torch.clamp(xqnorm, 0.0, 1.0)
    else:
        raise RuntimeError("get_extrap_pos only work for periodic and mirror extrapolation")
    return xqinside * (xmax - xmin) + xmin

def get_extrap_val(xqextrap, y, extrap):
    # xqextrap: (nrq,)
    # y: (*BY, nr)
    shape = (*y.shape[:-1], xqextrap.shape[-1])
    dtype = xqextrap.dtype
    device = xqextrap.device

    if extrap is None or extrap == "nan":
        return torch.empty(shape, dtype=dtype, device=device) * float("nan")
    elif isinstance(extrap, int) or isinstance(extrap, float) or \
            (isinstance(extrap, torch.Tensor) and torch.numel(extrap) == 1):
        return torch.zeros(shape, dtype=dtype, device=device) + extrap
    elif hasattr(extrap, "__call__"):
        return extrap(xqextrap).expand(*y.shape[:-1], -1)  # (*BY, nrq)
    else:
        raise RuntimeError("Invalid extrap type (type: %s): %s" % (type(extrap), extrap))



import contextlib
import torch
import copy
from typing import Mapping, Callable, Union, Dict, List

def set_default_option(defopt: Dict, opt: Dict) -> Dict:
    # return a dictionary based on the options and if no item from option,
    # take it from defopt

    # make a shallow copy to detach the results from defopt
    res = copy.copy(defopt)
    res.update(opt)
    return res

def get_and_pop_keys(dct: Dict, keys: List) -> Dict:
    res = {}
    for k in keys:
        res[k] = dct.pop(k)
    return res

def get_method(
        algname: str,
        methods: Mapping[str, Callable],
        method: Union[str, Callable]) -> Callable:

    if isinstance(method, str):
        methodname = method.lower()
        if methodname in methods:
            return methods[methodname]
        else:
            raise RuntimeError("Unknown %s method: %s" % (algname, method))
    elif hasattr(method, "__call__"):
        return method
    elif method is None:
        assert False, "Internal assert failed, method in get_method is not supposed" \
            "to be None. If this shows, then the corresponding function fails to " \
            "set the default method"
    else:
        raise TypeError("Invalid method type: %s. Only str and callable are accepted." % type(method))

@contextlib.contextmanager
def dummy_context_manager():
    yield None

class TensorNonTensorSeparator(object):
    """
    Class that provides function to separate/combine tensors and nontensors
    parameters.
    """

    def __init__(self, params, varonly=True):
        """
        Params is a list of tensor or non-tensor to be splitted into
        tensor/non-tensor
        """
        self.tensor_idxs = []
        self.tensor_params = []
        self.nontensor_idxs = []
        self.nontensor_params = []
        self.nparams = len(params)
        for (i, p) in enumerate(params):
            if isinstance(p, torch.Tensor) and ((varonly and p.requires_grad) or (not varonly)):
                self.tensor_idxs.append(i)
                self.tensor_params.append(p)
            else:
                self.nontensor_idxs.append(i)
                self.nontensor_params.append(p)
        self.alltensors = len(self.tensor_idxs) == self.nparams

    def get_tensor_params(self):
        return self.tensor_params

    def ntensors(self):
        return len(self.tensor_idxs)

    def nnontensors(self):
        return len(self.nontensor_idxs)

    def reconstruct_params(self, tensor_params, nontensor_params=None):
        if nontensor_params is None:
            nontensor_params = self.nontensor_params
        if len(tensor_params) + len(nontensor_params) != self.nparams:
            raise ValueError(
                "The total length of tensor and nontensor params "
                "do not match with the expected length: %d instead of %d" %
                (len(tensor_params) + len(nontensor_params), self.nparams))
        if self.alltensors:
            return tensor_params

        params = [None for _ in range(self.nparams)]
        for nidx, p in zip(self.nontensor_idxs, nontensor_params):
            params[nidx] = p
        for idx, p in zip(self.tensor_idxs, tensor_params):
            params[idx] = p
        return params

class TensorPacker(object):
    def __init__(self, tensors):
        self.idx_shapes = []
        istart = 0
        for i, p in enumerate(tensors):
            ifinish = istart + torch.numel(p)
            self.idx_shapes.append((istart, ifinish, p.shape))
            istart = ifinish

    def flatten(self, y_list):
        return torch.cat([y.reshape(-1) for y in y_list], dim=-1)

    def pack(self, y):
        yshapem1 = y.shape[:-1]
        return tuple(
            y[..., istart:ifinish].reshape(yshapem1 + shape)
            for (istart, ifinish, shape) in self.idx_shapes
        )



from typing import List, Optional, Dict
from xitorch._utils.assertfuncs import assert_runtime

class Uniquifier(object):
    def __init__(self, allobjs: List):
        self.nobjs = len(allobjs)

        id2idx: Dict[int, int] = {}
        unique_objs: List[int] = []
        unique_idxs: List[int] = []
        nonunique_map_idxs: List[int] = [-self.nobjs * 2] * self.nobjs
        num_unique = 0
        for i, obj in enumerate(allobjs):
            id_obj = id(obj)
            if id_obj in id2idx:
                nonunique_map_idxs[i] = id2idx[id_obj]
                continue
            id2idx[id_obj] = num_unique
            unique_objs.append(obj)
            nonunique_map_idxs[i] = num_unique
            unique_idxs.append(i)
            num_unique += 1

        self.unique_objs = unique_objs
        self.unique_idxs = unique_idxs
        self.nonunique_map_idxs = nonunique_map_idxs
        self.num_unique = num_unique
        self.all_unique = self.nobjs == self.num_unique

    def get_unique_objs(self, allobjs: Optional[List] = None) -> List:
        if allobjs is None:
            return self.unique_objs
        assert_runtime(len(allobjs) == self.nobjs, "The allobjs must have %d elements" % self.nobjs)
        if self.all_unique:
            return allobjs
        return [allobjs[i] for i in self.unique_idxs]

    def map_unique_objs(self, uniqueobjs: List) -> List:
        assert_runtime(len(uniqueobjs) == self.num_unique, "The uniqueobjs must have %d elements" % self.num_unique)
        if self.all_unique:
            return uniqueobjs
        return [uniqueobjs[idx] for idx in self.nonunique_map_idxs]



import torch
import functools
from typing import Optional

__all__ = ["rk23_adaptive", "rk45_adaptive"]

def rk_step(func, t, y, f, h, abck):
    # A: (norder, norder)
    # B: (norder,)
    # C: (norder,)
    # K: (norder+1, ny)
    A, B, C, K = abck
    K[0] = f
    for s, (a, c) in enumerate(zip(A[1:], C[1:]), start=1):
        dy = torch.matmul(K[:s].T, a[:s]) * h
        K[s] = func(t + c * h, y + dy)
    ynew = y + h * torch.matmul(K[:-1].T, B)
    fnew = func(t + h, ynew)
    K[-1] = fnew
    return ynew, fnew

class RKAdaptiveStepSolver(object):
    A: Optional[torch.Tensor] = None
    B: Optional[torch.Tensor] = None
    C: Optional[torch.Tensor] = None
    E: Optional[torch.Tensor] = None
    n_stages: Optional[int] = None
    error_estimator_order: Optional[int] = None

    def __init__(self, atol, rtol):
        self.atol = atol
        self.rtol = rtol
        self.max_factor = 10
        self.min_factor = 0.2
        self.step_mult = 0.9
        self.error_exponent = -1. / (self.error_estimator_order + 1.)

    def setup(self, fcn, ts, y0, params):
        # flatten the y0, will be restore at the end of .solve()
        self.yshape = y0.shape
        self.y0 = y0.reshape(-1)

        direction = ts[1] - ts[0]
        if direction < 0:
            self.ts = -ts
            self.func = lambda t, y: -fcn(-t, y.reshape(self.yshape), *params).reshape(-1)
        else:
            self.ts = ts
            self.func = lambda t, y: fcn(t, y.reshape(self.yshape), *params).reshape(-1)
        self.dtype = y0.dtype
        self.device = y0.device
        n = torch.numel(y0)
        self.K = torch.empty((self.n_stages + 1, n), dtype=self.dtype, device=self.device)

        # convert the predefined tensors into the dtype and device
        self.A = self.A.to(self.dtype).to(self.device)
        self.B = self.B.to(self.dtype).to(self.device)
        self.C = self.C.to(self.dtype).to(self.device)
        self.E = self.E.to(self.dtype).to(self.device)

    def solve(self):
        t0 = self.ts[0]
        ts = self.ts
        f0 = self.func(t0, self.y0)
        h0 = self.ts[1] - self.ts[0]  # ??? perform more intelligent guess

        # prepare the results
        nt = len(ts)
        yt = torch.empty((len(self.ts), *self.y0.shape), dtype=self.dtype, device=self.device)
        yt[0] = self.y0

        rk_state = (f0, t0, self.y0, h0)
        for i in range(1, len(ts)):
            rk_state = self._step(rk_state, ts[i])
            yt[i] = rk_state[2]
        return yt.reshape(-1, *self.yshape)

    def _error_norm(self, K, h):
        err = torch.matmul(K.T, self.E) * h
        return err.norm()

    def _step(self, rk_state, t1):
        t1_achieved = False
        while not t1_achieved:
            rk_state, t1_achieved = self._single_step(rk_state, t1)
        return rk_state

    def _single_step(self, rk_state, t1):
        f0, t0, y0, h = rk_state
        accepted = False
        prev_rejected = False
        while not accepted:
            # check if the current step exceeds the target
            t1_achieved = t0 + h > t1
            hstep = t1 - t0 if t1_achieved else h
            tnew = t0 + hstep

            # perform the RK-step to t0+h
            abck = (self.A, self.B, self.C, self.K)
            ynew, fnew = rk_step(self.func, t0, y0, f0, hstep, abck)

            # estimate the error norm
            scale = self.atol + torch.max(y0.norm(), ynew.norm()) * self.rtol
            errnorm = self._error_norm(self.K, hstep) / scale
            accepted = errnorm < 1

            # adjust the step size
            if accepted and not t1_achieved:
                if errnorm == 0:
                    factor = self.max_factor
                else:
                    factor = min(self.max_factor, self.step_mult * errnorm ** self.error_exponent)

                if prev_rejected:
                    factor = min(1.0, factor)

                h *= factor
            elif not accepted:
                factor = max(self.min_factor, self.step_mult * errnorm ** self.error_exponent)
                h = hstep * factor

            prev_rejected = not accepted

        rk_state = (fnew, tnew, ynew, h)
        return rk_state, t1_achieved

class RK23(RKAdaptiveStepSolver):
    error_estimator_order = 2
    n_stages = 3
    C = torch.tensor([0, 1 / 2, 3 / 4], dtype=torch.float64)
    A = torch.tensor([
        [0, 0, 0],
        [1 / 2, 0, 0],
        [0, 3 / 4, 0]
    ], dtype=torch.float64)
    B = torch.tensor([2 / 9, 1 / 3, 4 / 9], dtype=torch.float64)
    E = torch.tensor([5 / 72, -1 / 12, -1 / 9, 1 / 8], dtype=torch.float64)

class RK45(RKAdaptiveStepSolver):
    error_estimator_order = 4
    n_stages = 6
    C = torch.tensor([0, 1 / 5, 3 / 10, 4 / 5, 8 / 9, 1], dtype=torch.float64)
    A = torch.tensor([
        [0, 0, 0, 0, 0],
        [1 / 5, 0, 0, 0, 0],
        [3 / 40, 9 / 40, 0, 0, 0],
        [44 / 45, -56 / 15, 32 / 9, 0, 0],
        [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729, 0],
        [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656]
    ], dtype=torch.float64)
    B = torch.tensor([35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84], dtype=torch.float64)
    E = torch.tensor([-71 / 57600, 0, 71 / 16695, -71 / 1920, 17253 / 339200, -22 / 525,
                      1 / 40], dtype=torch.float64)

def _rk_adaptive(fcn, ts, y0, params, cls, atol=1e-8, rtol=1e-5, **unused):
    """
    Keyword arguments
    -----------------
    atol: float
        The absolute error tolerance in deciding the steps
    rtol: float
        The relative error tolerance in deciding the steps
    """
    solver = cls(atol=atol, rtol=rtol)
    solver.setup(fcn, ts, y0, params)
    return solver.solve()

@functools.wraps(_rk_adaptive, assigned='__annotations__')
def rk23_adaptive(fcn, ts, y0, params, **kwargs):
    """
    Perform the adaptive Runge-Kutta steps with order 2 and 3.
    """
    return _rk_adaptive(fcn, ts, y0, params, RK23, **kwargs)

@functools.wraps(_rk_adaptive, assigned='__annotations__')
def rk45_adaptive(fcn, ts, y0, params, **kwargs):
    """
    Perform the adaptive Runge-Kutta steps with order 4 and 5.
    """
    return _rk_adaptive(fcn, ts, y0, params, RK45, **kwargs)


# complete the docstring
rk23_adaptive.__doc__ += _rk_adaptive.__doc__  # type: ignore
rk45_adaptive.__doc__ += _rk_adaptive.__doc__  # type: ignore



import torch
from xitorch._core.editable_module import EditableModule
from xitorch._impls.integrate.samples_quad import CubicSplineSQuad, TrapzSQuad, SimpsonSQuad
from xitorch._utils.misc import get_method
from xitorch._docstr.api_docstr import get_methods_docstr
from typing import List, Union, Callable

__all__ = ["SQuad"]

class SQuad(EditableModule):
    r"""
    SQuad (Sampled QUADrature) is a class for quadrature performed with a
    fixed samples at given points.
    Mathematically, it does the integration

    .. math::

        \mathbf{z}(x) = \int_{x_0}^x \mathbf{y}(x')\ \mathrm{d}x

    where :math:`\mathbf{y}(x)` is the interpolated function from a given sample.

    Arguments
    ---------
    x: torch.Tensor
        The positions where the samples are given. It is a 1D tensor with shape
        ``(nx,)``.
    method: str or callable or None
        The integration method. If None, it will choose ``"cspline"``.
    **fwd_options
        Method-specific options (see method section below)
    """

    def __init__(self, x: torch.Tensor,
                 method: Union[str, Callable, None] = None,
                 **fwd_options):
        if method is None:
            method = "cspline"
        if not (isinstance(x, torch.Tensor) and len(x.shape) == 1):
            raise RuntimeError("The input x to SQuad must be a 1D tensor")

        all_clss = {
            "cspline": CubicSplineSQuad,
            "simpson": SimpsonSQuad,
            "trapz": TrapzSQuad,
        }
        clss = get_method("SQuad", all_clss, method)
        self.obj = clss(x, **fwd_options)
        self.nx = x.shape[-1]

    def cumsum(self, y: torch.Tensor, dim: int = -1) -> torch.Tensor:
        r"""
        Perform the cumulative integration of the samples :math:`\mathbf{y}`
        over the specified dimension.

        Arguments
        ---------
        y: torch.Tensor
            The value of samples. The size of ``y`` at ``dim`` must be equal
            to the length of ``x``.
        dim: int
            The dimension where the cumulative integration is performed.

        Returns
        -------
        torch.Tensor
            The cumulative integrated values with the same shape as ``y``.
        """
        swapaxes = dim != -1
        if swapaxes:
            y = y.transpose(dim, -1)
        if y.shape[-1] != self.nx:
            raise RuntimeError("The length of integrated dimension does not match with x")
        res = self.obj.cumsum(y)
        if swapaxes:
            res = res.transpose(dim, -1)
        return res

    def integrate(self, y: torch.Tensor, dim: int = -1, keepdim: bool = False) -> torch.Tensor:
        r"""
        Perform the full integration of the samples :math:`\mathbf{y}`
        over the specified dimension.

        Arguments
        ---------
        y: torch.Tensor
            The value of samples. The size of ``y`` at ``dim`` must be equal
            to the length of ``x``, i.e. ``(..., nx, ...)``.
        dim: int
            The dimension where the integration is performed.
        keepdim: bool
            Option to not discard the integrated dimension. If ``True``, the
            integrated dimension size will be 1.

        Returns
        -------
        torch.Tensor
            The integrated values.
        """
        swapaxes = dim != -1
        if swapaxes:
            y = y.transpose(dim, -1)
        if y.shape[-1] != self.nx:
            raise RuntimeError("The length of integrated dimension does not match with x")
        res = self.obj.integrate(y)
        if keepdim:
            res = res.unsqueeze(-1)
        if swapaxes:
            res = res.transpose(dim, -1)
        return res

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """"""
        return self.obj.getparamnames(methodname, prefix=prefix + "obj.")


# docstring completion
_squad_methods = {
    "cspline": CubicSplineSQuad,
    # "simpson": SimpsonSQuad,
    "trapz": TrapzSQuad,
}
SQuad.__doc__ = get_methods_docstr(SQuad, _squad_methods)



from typing import List, Callable, Sequence, NamedTuple, Union
import torch

# All functions in this file should have the following inputs and outputs
# Inputs
# ------
# * fcn: callable dy/dt = fcn(t, y, *params)
#       The function to be integrated. It should produce output of list of
#       tensors following the shapes of tuple `y`. `t` should be a single element.
# * t: torch.Tensor (nt,)
#       The integrated times
# * y0: list of torch.Tensor (*ny)
#       The list of initial values
# * params: list
#       List of any other parameters
# * **kwargs: dict
#       Any other keyword arguments
# Outputs
# -------
# * yt: list of torch.Tensor (nt,*ny)
#       The value of `y` at the given time `t`
# Note
# ----
# The operations are done in grad-disabled environment and **not** expected to
# be able to propagate gradients.

__all__ = ["rk4_ivp", "rk38_ivp", "fwd_euler_ivp"]

############################# list of tableaus #############################
class _Tableau(NamedTuple):
    c: List[float]
    b: List[float]
    a: List[List[float]]

rk4_tableau = _Tableau(
    c=[0.0, 0.5, 0.5, 1.0],
    b=[1 / 6., 1 / 3., 1 / 3., 1 / 6.],
    a=[[0.0, 0.0, 0.0, 0.0],
       [0.5, 0.0, 0.0, 0.0],
       [0.0, 0.5, 0.0, 0.0],
       [0.0, 0.0, 1.0, 0.0]]
)
rk38_tableau = _Tableau(
    c=[0.0, 1 / 3, 2 / 3, 1.0],
    b=[1 / 8, 3 / 8, 3 / 8, 1 / 8],
    a=[[0.0, 0.0, 0.0, 0.0],
       [1 / 3, 0.0, 0.0, 0.0],
       [-1 / 3, 1.0, 0.0, 0.0],
       [1.0, -1.0, 1.0, 0.0]]
)
fwd_euler_tableau = _Tableau(
    c=[0.0],
    b=[1.0],
    a=[[0.0]]
)

def explicit_rk(tableau: _Tableau,
                fcn: Callable[..., torch.Tensor], t: torch.Tensor, y0: torch.Tensor,
                params: Sequence[torch.Tensor]):
    c = tableau.c
    a = tableau.a
    b = tableau.b
    s = len(c)
    nt = len(t)
    dtype = t.dtype
    device = t.device

    # set up the results list
    yt_lst: List[torch.Tensor] = []
    yt_lst.append(y0)
    y = y0
    # see https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods#Explicit_Runge.E2.80.93Kutta_methods
    # for the implementation
    for i in range(nt - 1):
        t0 = t[i]
        t1 = t[i + 1]
        h = t1 - t0
        ks: List[torch.Tensor] = []
        ksum: Union[float, torch.Tensor] = 0.0
        for j in range(s):
            if j == 0:
                k = fcn(t0, y, *params)
            else:
                ak: Union[float, torch.Tensor] = 0.0
                aj = a[j]
                for m in range(j):
                    ak = aj[m] * ks[m] + ak
                k = fcn(t0 + c[j] * h, h * ak + y, *params)
            ks.append(k)
            ksum = ksum + b[j] * k
        y = h * ksum + y
        yt_lst.append(y)
    yt = torch.stack(yt_lst, dim=0)
    return yt

############################# list of methods #############################
def rk38_ivp(fcn: Callable[..., torch.Tensor], t: torch.Tensor, y0: torch.Tensor,
             params: Sequence[torch.Tensor], **kwargs):
    return explicit_rk(rk38_tableau, fcn, t, y0, params)

def fwd_euler_ivp(fcn: Callable[..., torch.Tensor], t: torch.Tensor, y0: torch.Tensor,
                  params: Sequence[torch.Tensor], **kwargs):
    return explicit_rk(fwd_euler_tableau, fcn, t, y0, params)

def rk4_ivp(fcn: Callable[..., torch.Tensor], t: torch.Tensor, y0: torch.Tensor,
            params: Sequence[torch.Tensor], **kwargs):
    """
    Perform the Runge-Kutta steps of order 4 with a fixed step size.
    """
    return explicit_rk(rk4_tableau, fcn, t, y0, params)



import torch
import warnings
from typing import Union, Any, Mapping, Optional, Callable
from xitorch import LinearOperator
from xitorch._core.linop import MatrixLinearOperator
from xitorch._utils.assertfuncs import assert_runtime
from xitorch._utils.misc import set_default_option, dummy_context_manager, get_method
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch.debug.modes import is_debug_enabled
from xitorch._impls.linalg.solve import exactsolve, wrap_gmres, \
    cg, bicgstab, broyden1_solve, _get_batchdims, gmres

def solve(A: LinearOperator, B: torch.Tensor, E: Union[torch.Tensor, None] = None,
          M: Optional[LinearOperator] = None,
          bck_options: Mapping[str, Any] = {},
          method: Union[str, Callable, None] = None,
          **fwd_options) -> torch.Tensor:
    r"""
    Performing iterative method to solve the equation

    .. math::

        \mathbf{AX=B}

    or

    .. math::

        \mathbf{AX-MXE=B}

    where :math:`\mathbf{E}` is a diagonal matrix.
    This function can also solve batched multiple inverse equation at the
    same time by applying :math:`\mathbf{A}` to a tensor :math:`\mathbf{X}`
    with shape ``(...,na,ncols)``.
    The applied :math:`\mathbf{E}` are not necessarily identical for each column.

    Arguments
    ---------
    A: xitorch.LinearOperator
        A linear operator that takes an input ``X`` and produce the vectors in the same
        space as ``B``.
        It should have the shape of ``(*BA, na, na)``
    B: torch.Tensor
        The tensor on the right hand side with shape ``(*BB, na, ncols)``
    E: torch.Tensor or None
        If a tensor, it will solve :math:`\mathbf{AX-MXE = B}`.
        It will be regarded as the diagonal of the matrix.
        Otherwise, it just solves :math:`\mathbf{AX = B}` and ``M`` is ignored.
        If it is a tensor, it should have shape of ``(*BE, ncols)``.
    M: xitorch.LinearOperator or None
        The transformation on the ``E`` side. If ``E`` is ``None``,
        then this argument is ignored.
        If E is not ``None`` and ``M`` is ``None``, then ``M=I``.
        If LinearOperator, it must be Hermitian with shape ``(*BM, na, na)``.
    bck_options: dict
        Options of the iterative solver in the backward calculation.
    method: str or callable or None
        The method of linear equation solver. If ``None``, it will choose
        ``"cg"`` or ``"bicgstab"`` based on the matrices symmetry.
        `Note`: default method will be changed quite frequently, so if you want
        future compatibility, please specify a method.
    **fwd_options
        Method-specific options (see method below)

    Returns
    -------
    torch.Tensor
        The tensor :math:`\mathbf{X}` that satisfies :math:`\mathbf{AX-MXE=B}`.
    """
    assert_runtime(A.shape[-1] == A.shape[-2], "The linear operator A must have a square shape")
    assert_runtime(A.shape[-1] == B.shape[-2], "Mismatch shape of A & B (A: %s, B: %s)" % (A.shape, B.shape))
    assert_runtime(
        not torch.is_grad_enabled() or A.is_getparamnames_implemented,
        "The _getparamnames(self, prefix) of linear operator A must be "
        "implemented if using solve with grad enabled")
    if M is not None:
        assert_runtime(M.shape[-1] == M.shape[-2], "The linear operator M must have a square shape")
        assert_runtime(M.shape[-1] == A.shape[-1], "The shape of A & M must match (A: %s, M: %s)" % (A.shape, M.shape))
        assert_runtime(M.is_hermitian, "The linear operator M must be a Hermitian matrix")
        assert_runtime(
            not torch.is_grad_enabled() or M.is_getparamnames_implemented,
            "The _getparamnames(self, prefix) of linear operator M must be "
            "implemented if using solve with grad enabled")
    if E is not None:
        assert_runtime(E.shape[-1] == B.shape[-1],
                       "The last dimension of E & B must match (E: %s, B: %s)" % (E.shape, B.shape))
    if E is None and M is not None:
        warnings.warn("M is supplied but will be ignored because E is not supplied")

    # perform expensive check if debug mode is enabled
    if is_debug_enabled():
        A.check()
        if M is not None:
            M.check()

    if method is None:
        if isinstance(A, MatrixLinearOperator) and \
           (M is None or isinstance(M, MatrixLinearOperator)):
            method = "exactsolve"
        elif A.shape[-1] <= 5:  # for small matrix
            method = "exactsolve"
        else:
            is_hermit = A.is_hermitian and (M is None or M.is_hermitian)
            method = "cg" if is_hermit else "bicgstab"

    if method == "exactsolve":
        return exactsolve(A, B, E, M)
    else:
        # get the unique parameters of A
        params = A.getlinopparams()
        mparams = M.getlinopparams() if M is not None else []
        na = len(params)
        return solve_torchfcn.apply(
            A, B, E, M, method,
            fwd_options, bck_options,
            na, *params, *mparams)

class solve_torchfcn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A, B, E, M, method,
                fwd_options, bck_options,
                na, *all_params):
        # A: (*BA, nr, nr)
        # B: (*BB, nr, ncols)
        # E: (*BE, ncols) or None
        # M: (*BM, nr, nr) or None
        # all_params: list of tensor of any shape
        # returns: (*BABEM, nr, ncols)

        # separate the parameters for A and for M
        params = all_params[:na]
        mparams = all_params[na:]

        config = set_default_option({
        }, fwd_options)
        ctx.bck_config = set_default_option({
        }, bck_options)

        if torch.all(B == 0):  # special case
            dims = (*_get_batchdims(A, B, E, M), *B.shape[-2:])
            x = torch.zeros(dims, dtype=B.dtype, device=B.device)
        else:
            with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
                methods = {
                    "custom_exactsolve": custom_exactsolve,
                    "scipy_gmres": wrap_gmres,
                    "broyden1": broyden1_solve,
                    "cg": cg,
                    "bicgstab": bicgstab,
                    "gmres": gmres,
                }
                method_fcn = get_method("solve", methods, method)
                x = method_fcn(A, B, E, M, **config)

        ctx.e_is_none = E is None
        ctx.A = A
        ctx.M = M
        if ctx.e_is_none:
            ctx.save_for_backward(x, *all_params)
        else:
            ctx.save_for_backward(x, E, *all_params)
        ctx.na = na
        return x

    @staticmethod
    def backward(ctx, grad_x):
        # grad_x: (*BABEM, nr, ncols)
        # x: (*BABEM, nr, ncols)
        x = ctx.saved_tensors[0]
        idx_all_params = 1 if ctx.e_is_none else 2
        all_params = ctx.saved_tensors[idx_all_params:]
        params = all_params[:ctx.na]
        mparams = all_params[ctx.na:]
        E = None if ctx.e_is_none else ctx.saved_tensors[1]

        # solve (A-biases*M)^T v = grad_x
        # this is the grad of B
        with ctx.A.uselinopparams(*params), \
             ctx.M.uselinopparams(*mparams) if ctx.M is not None else dummy_context_manager():
            AT = ctx.A.H  # (*BA, nr, nr)
            MT = ctx.M.H if ctx.M is not None else None  # (*BM, nr, nr)
            Econj = E.conj() if E is not None else None
            v = solve(AT, grad_x, Econj, MT,
                      bck_options=ctx.bck_config, **ctx.bck_config)  # (*BABEM, nr, ncols)
        grad_B = v

        # calculate the grad of matrices parameters
        with torch.enable_grad():
            params = [p.clone().requires_grad_() for p in params]
            with ctx.A.uselinopparams(*params):
                loss = -ctx.A.mm(x)  # (*BABEM, nr, ncols)

        grad_params = torch.autograd.grad((loss,), params, grad_outputs=(v,),
                                          create_graph=torch.is_grad_enabled(),
                                          allow_unused=True)

        # calculate the biases gradient
        grad_E = None
        if E is not None:
            if ctx.M is None:
                Mx = x
            else:
                with ctx.M.uselinopparams(*mparams):
                    Mx = ctx.M.mm(x)  # (*BABEM, nr, ncols)
            grad_E = torch.einsum('...rc,...rc->...c', v, Mx.conj())  # (*BABEM, ncols)

        # calculate the gradient to the biases matrices
        grad_mparams = []
        if ctx.M is not None and E is not None:
            with torch.enable_grad():
                mparams = [p.clone().requires_grad_() for p in mparams]
                lmbdax = x * E.unsqueeze(-2)
                with ctx.M.uselinopparams(*mparams):
                    mloss = ctx.M.mm(lmbdax)

            grad_mparams = torch.autograd.grad((mloss,), mparams,
                                               grad_outputs=(v,),
                                               create_graph=torch.is_grad_enabled(),
                                               allow_unused=True)

        return (None, grad_B, grad_E, None, None, None, None, None,
                *grad_params, *grad_mparams)

def custom_exactsolve(A, B, E=None,
                      M=None, **options):
    # A: (*BA, na, na)
    # B: (*BB, na, ncols)
    # E: (*BE, ncols)
    # M: (*BM, na, na)
    return exactsolve(A, B, E, M)


# docstring completion
_solve_methods = {
    "cg": cg,
    "bicgstab": bicgstab,
    "exactsolve": exactsolve,
    "broyden1": broyden1_solve,
    "scipy_gmres": wrap_gmres,
    "gmres": gmres,
}
ignore_kwargs = ["E", "M", "mparams"]
solve.__doc__ = get_methods_docstr(solve, _solve_methods, ignore_kwargs)



import inspect
from xitorch._core.editable_module import EditableModule

def assert_broadcastable(shape1, shape2):
    if len(shape1) > len(shape2):
        assert_broadcastable(shape2, shape1)
        return
    for a, b in zip(shape1[::-1], shape2[::-1][:len(shape1)]):
        assert (a == 1 or b == 1 or a == b), "The shape %s and %s are not broadcastable" % (shape1, shape2)

def assert_fcn_params(fcn, args):
    if inspect.ismethod(fcn) and isinstance(fcn.__self__, EditableModule):
        fcn.__self__.assertparams(fcn, *args)

def assert_runtime(cond, msg=""):
    if not cond:
        raise RuntimeError(msg)

def assert_type(cond, msg=""):
    if not cond:
        raise TypeError(msg)



from abc import abstractmethod

class BaseInterp(object):
    @abstractmethod
    def __init__(self, x, y=None, **kwargs):
        pass

    @abstractmethod
    def __call__(self, xq, y=None):
        pass

    @abstractmethod
    def getparamnames(self):
        pass



from xitorch._core.editable_module import *
from xitorch._core.pure_function import *
from xitorch._core.linop import *
from xitorch._core.packer import *
from xitorch.debug.modes import *
from xitorch._utils.exceptions import *

from xitorch.version import get_version as _get_version
__version__ = _get_version()



import warnings
import torch
from typing import Union, Mapping, Any, Optional, Tuple, Callable
from xitorch import LinearOperator
from xitorch._core.linop import MatrixLinearOperator
from xitorch.linalg.solve import solve
from xitorch.debug.modes import is_debug_enabled
from xitorch._utils.assertfuncs import assert_runtime
from xitorch._utils.misc import set_default_option, \
    dummy_context_manager, get_method, get_and_pop_keys
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch._impls.linalg.symeig import exacteig, davidson
from xitorch._utils.exceptions import MathWarning

__all__ = ["lsymeig", "usymeig", "symeig", "svd"]

def lsymeig(A: LinearOperator, neig: Optional[int] = None,
            M: Optional[LinearOperator] = None,
            bck_options: Mapping[str, Any] = {},
            method: Union[str, Callable, None] = None,
            **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    return symeig(A, neig, "lowest", M, method=method, bck_options=bck_options, **fwd_options)

def usymeig(A: LinearOperator, neig: Optional[int] = None,
            M: Optional[LinearOperator] = None,
            bck_options: Mapping[str, Any] = {},
            method: Union[str, Callable, None] = None,
            **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    return symeig(A, neig, "uppest", M, method=method, bck_options=bck_options, **fwd_options)

def symeig(A: LinearOperator, neig: Optional[int] = None,
           mode: str = "lowest", M: Optional[LinearOperator] = None,
           bck_options: Mapping[str, Any] = {},
           method: Union[str, Callable, None] = None,
           **fwd_options) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""
    Obtain ``neig`` lowest eigenvalues and eigenvectors of a linear operator,

    .. math::

        \mathbf{AX = MXE}

    where :math:`\mathbf{A}, \mathbf{M}` are linear operators,
    :math:`\mathbf{E}` is a diagonal matrix containing the eigenvalues, and
    :math:`\mathbf{X}` is a matrix containing the eigenvectors.
    This function can handle derivatives for degenerate cases by setting non-zero
    ``degen_atol`` and ``degen_rtol`` in the backward option using the expressions
    in [1]_.

    Arguments
    ---------
    A: xitorch.LinearOperator
        The linear operator object on which the eigenpairs are constructed.
        It must be a Hermitian linear operator with shape ``(*BA, q, q)``
    neig: int or None
        The number of eigenpairs to be retrieved. If ``None``, all eigenpairs are
        retrieved
    mode: str
        ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
        it will take the lowest ``neig`` eigenpairs.
        If ``"uppest"``, it will take the uppermost ``neig``.
    M: xitorch.LinearOperator
        The transformation on the right hand side. If ``None``, then ``M=I``.
        If specified, it must be a Hermitian with shape ``(*BM, q, q)``.
    bck_options: dict
        Method-specific options for :func:`solve` which used in backpropagation
        calculation with some additional arguments for computing the backward
        derivatives:

        * ``degen_atol`` (``float`` or None): Minimum absolute difference between
          two eigenvalues to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.6``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)
        * ``degen_rtol`` (``float`` or None): Minimum relative difference between
          two eigenvalues to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.4``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)

        Note: the default values of ``degen_atol`` and ``degen_rtol`` are going
        to change in the future. So, for future compatibility, please specify
        the specific values.

    method: str or callable or None
        Method for the eigendecomposition. If ``None``, it will choose
        ``"exacteig"``.
    **fwd_options
        Method-specific options (see method section below).

    Returns
    -------
    tuple of tensors (eigenvalues, eigenvectors)
        It will return eigenvalues and eigenvectors with shapes respectively
        ``(*BAM, neig)`` and ``(*BAM, na, neig)``, where ``*BAM`` is the
        broadcasted shape of ``*BA`` and ``*BM``.

    References
    ----------
    .. [1] Muhammad F. Kasim,
           "Derivatives of partial eigendecomposition of a real symmetric matrix for degenerate cases".
           arXiv:2011.04366 (2020)
           `https://arxiv.org/abs/2011.04366 <https://arxiv.org/abs/2011.04366>`_
    """
    assert_runtime(A.is_hermitian, "The linear operator A must be Hermitian")
    assert_runtime(
        not torch.is_grad_enabled() or A.is_getparamnames_implemented,
        "The _getparamnames(self, prefix) of linear operator A must be "
        "implemented if using symeig with grad enabled")
    if M is not None:
        assert_runtime(M.is_hermitian, "The linear operator M must be Hermitian")
        assert_runtime(M.shape[-1] == A.shape[-1], "The shape of A & M must match (A: %s, M: %s)" % (A.shape, M.shape))
        assert_runtime(
            not torch.is_grad_enabled() or M.is_getparamnames_implemented,
            "The _getparamnames(self, prefix) of linear operator M must be "
            "implemented if using symeig with grad enabled")
    mode = mode.lower()
    if mode == "uppermost":
        mode = "uppest"
    if method is None:
        if isinstance(A, MatrixLinearOperator) and \
           (M is None or isinstance(M, MatrixLinearOperator)):
            method = "exacteig"
        else:
            # TODO: implement robust LOBPCG and put it here
            method = "exacteig"
    if neig is None:
        neig = A.shape[-1]

    # perform expensive check if debug mode is enabled
    if is_debug_enabled():
        A.check()
        if M is not None:
            M.check()

    if method == "exacteig":
        return exacteig(A, neig, mode, M)
    else:
        fwd_options["method"] = method
        # get the unique parameters of A & M
        params = A.getlinopparams()
        mparams = M.getlinopparams() if M is not None else []
        na = len(params)
        return symeig_torchfcn.apply(A, neig, mode, M,
                                     fwd_options, bck_options,
                                     na, *params, *mparams)

def svd(A: LinearOperator, k: Optional[int] = None,
        mode: str = "uppest", bck_options: Mapping[str, Any] = {},
        method: Union[str, Callable, None] = None,
        **fwd_options) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    Perform the singular value decomposition (SVD):

    .. math::

        \mathbf{A} = \mathbf{U\Sigma V}^H

    where :math:`\mathbf{U}` and :math:`\mathbf{V}` are semi-unitary matrix and
    :math:`\mathbf{\Sigma}` is a diagonal matrix containing real non-negative
    numbers.
    This function can handle derivatives for degenerate singular values by setting non-zero
    ``degen_atol`` and ``degen_rtol`` in the backward option using the expressions
    in [1]_.

    Arguments
    ---------
    A: xitorch.LinearOperator
        The linear operator to be decomposed. It has a shape of ``(*BA, m, n)``
        where ``(*BA)`` is the batched dimension of ``A``.
    k: int or None
        The number of decomposition obtained. If ``None``, it will be
        ``min(*A.shape[-2:])``
    mode: str
        ``"lowest"`` or ``"uppermost"``/``"uppest"``. If ``"lowest"``,
        it will take the lowest ``k`` decomposition.
        If ``"uppest"``, it will take the uppermost ``k``.
    bck_options: dict
        Method-specific options for :func:`solve` which used in backpropagation
        calculation with some additional arguments for computing the backward
        derivatives:

        * ``degen_atol`` (``float`` or None): Minimum absolute difference between
          two singular values to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.6``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)
        * ``degen_rtol`` (``float`` or None): Minimum relative difference between
          two singular values to be treated as degenerate. If None, it is
          ``torch.finfo(dtype).eps**0.4``. If 0.0, no special treatment on
          degeneracy is applied. (default: None)

        Note: the default values of ``degen_atol`` and ``degen_rtol`` are going
        to change in the future. So, for future compatibility, please specify
        the specific values.

    method: str or callable or None
        Method for the svd (same options for :func:`symeig`). If ``None``,
        it will choose ``"exacteig"``.
    **fwd_options
        Method-specific options (see method section below).

    Returns
    -------
    tuple of tensors (u, s, vh)
        It will return ``u, s, vh`` with shapes respectively
        ``(*BA, m, k)``, ``(*BA, k)``, and ``(*BA, k, n)``.

    Note
    ----
    It is a naive implementation of symmetric eigendecomposition of ``A.H @ A``
    or ``A @ A.H`` (depending which one is cheaper)

    References
    ----------
    .. [1] Muhammad F. Kasim,
           "Derivatives of partial eigendecomposition of a real symmetric matrix for degenerate cases".
           arXiv:2011.04366 (2020)
           `https://arxiv.org/abs/2011.04366 <https://arxiv.org/abs/2011.04366>`_
    """
    # A: (*BA, m, n)
    # adapted from scipy.sparse.linalg.svds

    if is_debug_enabled():
        A.check()
    BA = A.shape[:-2]

    m = A.shape[-2]
    n = A.shape[-1]
    if m < n:
        AAsym = A.matmul(A.H, is_hermitian=True)
        min_nm = m
    else:
        AAsym = A.H.matmul(A, is_hermitian=True)
        min_nm = n

    eivals, eivecs = symeig(AAsym, k, mode,
                            bck_options=bck_options, method=method,
                            **fwd_options)  # (*BA, k) and (*BA, min(mn), k)

    # clamp the eigenvalues to a small positive values to avoid numerical
    # instability
    eivals = torch.clamp(eivals, min=0.0)
    s = torch.sqrt(eivals)  # (*BA, k)
    sdiv = torch.clamp(s, min=1e-12).unsqueeze(-2)  # (*BA, 1, k)
    if m < n:
        u = eivecs  # (*BA, m, k)
        v = A.rmm(u) / sdiv  # (*BA, n, k)
    else:
        v = eivecs  # (*BA, n, k)
        u = A.mm(v) / sdiv  # (*BA, m, k)
    vh = v.transpose(-2, -1).conj()
    return u, s, vh

class symeig_torchfcn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A, neig, mode, M, fwd_options, bck_options, na, *amparams):
        # A: LinearOperator (*BA, q, q)
        # M: LinearOperator (*BM, q, q) or None

        # separate the sets of parameters
        params = amparams[:na]
        mparams = amparams[na:]

        config = set_default_option({
        }, fwd_options)
        ctx.bck_config = set_default_option({
            "degen_atol": None,
            "degen_rtol": None,
        }, bck_options)

        # options for calculating the backward (not for `solve`)
        alg_keys = ["degen_atol", "degen_rtol"]
        ctx.bck_alg_config = get_and_pop_keys(ctx.bck_config, alg_keys)

        method = config.pop("method")
        with A.uselinopparams(*params), M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
            methods = {
                "davidson": davidson,
                "custom_exacteig": custom_exacteig,
            }
            method_fcn = get_method("symeig", methods, method)
            evals, evecs = method_fcn(A, neig, mode, M, **config)

        # save for the backward
        # evals: (*BAM, neig)
        # evecs: (*BAM, na, neig)
        ctx.save_for_backward(evals, evecs, *amparams)
        ctx.na = na
        ctx.A = A
        ctx.M = M
        return evals, evecs

    @staticmethod
    def backward(ctx, grad_evals, grad_evecs):
        # grad_evals: (*BAM, neig)
        # grad_evecs: (*BAM, na, neig)

        # get the variables from ctx
        evals, evecs = ctx.saved_tensors[:2]
        na = ctx.na
        amparams = ctx.saved_tensors[2:]
        params = amparams[:na]
        mparams = amparams[na:]

        M = ctx.M
        A = ctx.A
        degen_atol: Optional[float] = ctx.bck_alg_config["degen_atol"]  # type: ignore
        degen_rtol: Optional[float] = ctx.bck_alg_config["degen_rtol"]  # type: ignore

        # set the default values of degen_*tol
        dtype = evals.dtype
        if degen_atol is None:
            degen_atol = torch.finfo(dtype).eps**0.6
        if degen_rtol is None:
            degen_rtol = torch.finfo(dtype).eps**0.4

        # check the degeneracy
        if degen_atol > 0 or degen_rtol > 0:
            # idx_degen: (*BAM, neig, neig)
            idx_degen, isdegenerate = _check_degen(evals, degen_atol, degen_rtol)
        else:
            isdegenerate = False
        if not isdegenerate:
            idx_degen = None

        # the loss function where the gradient will be retrieved
        # warnings: if not all params have the connection to the output of A,
        # it could cause an infinite loop because pytorch will keep looking
        # for the *params node and propagate further backward via the `evecs`
        # path. So make sure all the *params are all connected in the graph.
        with torch.enable_grad():
            params = [p.clone().requires_grad_() for p in params]
            with A.uselinopparams(*params):
                loss = A.mm(evecs)  # (*BAM, na, neig)

        # if degenerate, check the conditions for finite derivative
        if is_debug_enabled() and isdegenerate:
            xtg = torch.matmul(evecs.transpose(-2, -1).conj(), grad_evecs)
            req1 = idx_degen * (xtg - xtg.transpose(-2, -1).conj())
            reqtol = xtg.abs().max() * grad_evecs.shape[-2] * torch.finfo(grad_evecs.dtype).eps

            if not torch.all(torch.abs(req1) <= reqtol):
                # if the requirements are not satisfied, raises a warning
                msg = ("Degeneracy appears but the loss function seem to depend "
                       "strongly on the eigenvector. The gradient might be incorrect.\n")
                msg += "Eigenvalues:\n%s\n" % str(evals)
                msg += "Degenerate map:\n%s\n" % str(idx_degen)
                msg += "Requirements (should be all 0s):\n%s" % str(req1)
                warnings.warn(MathWarning(msg))

        # calculate the contributions from the eigenvalues
        gevalsA = grad_evals.unsqueeze(-2) * evecs  # (*BAM, na, neig)

        # calculate the contributions from the eigenvectors
        with M.uselinopparams(*mparams) if M is not None else dummy_context_manager():
            # orthogonalize the grad_evecs with evecs
            B = _ortho(grad_evecs, evecs, D=idx_degen, M=M, mright=False)

            # Based on test cases, complex datatype is more likely to suffer from
            # singularity error when doing the inverse. Therefore, I add a small
            # offset here to prevent that from happening
            if torch.is_complex(B):
                evals_offset = evals + 1e-14
            else:
                evals_offset = evals

            with A.uselinopparams(*params):
                gevecs = solve(A, -B, evals_offset, M, bck_options=ctx.bck_config,
                               **ctx.bck_config)  # (*BAM, na, neig)

            # orthogonalize gevecs w.r.t. evecs
            gevecsA = _ortho(gevecs, evecs, D=None, M=M, mright=True)

        # accummulate the gradient contributions
        gaccumA = gevalsA + gevecsA
        grad_params = torch.autograd.grad(
            outputs=(loss,),
            inputs=params,
            grad_outputs=(gaccumA,),
            create_graph=torch.is_grad_enabled(),
        )

        grad_mparams = []
        if ctx.M is not None:
            with torch.enable_grad():
                mparams = [p.clone().requires_grad_() for p in mparams]
                with M.uselinopparams(*mparams):
                    mloss = M.mm(evecs)  # (*BAM, na, neig)
            gevalsM = -gevalsA * evals.unsqueeze(-2)
            gevecsM = -gevecsA * evals.unsqueeze(-2)

            # the contribution from the parallel elements
            gevecsM_par = (-0.5 * torch.einsum("...ae,...ae->...e", grad_evecs, evecs.conj())
                           ).unsqueeze(-2) * evecs  # (*BAM, na, neig)

            gaccumM = gevalsM + gevecsM + gevecsM_par
            grad_mparams = torch.autograd.grad(
                outputs=(mloss,),
                inputs=mparams,
                grad_outputs=(gaccumM,),
                create_graph=torch.is_grad_enabled(),
            )

        return (None, None, None, None, None, None, None, *grad_params, *grad_mparams)

def _check_degen(evals: torch.Tensor, degen_atol: float, degen_rtol: float) -> \
        Tuple[torch.Tensor, bool]:
    # evals: (*BAM, neig)

    # get the index of degeneracies
    neig = evals.shape[-1]
    evals_diff = torch.abs(evals.unsqueeze(-2) - evals.unsqueeze(-1))  # (*BAM, neig, neig)
    degen_thrsh = degen_atol + degen_rtol * torch.abs(evals).unsqueeze(-1)
    idx_degen = (evals_diff < degen_thrsh).to(evals.dtype)
    isdegenerate = bool(torch.sum(idx_degen) > torch.numel(evals))
    return idx_degen, isdegenerate

def _ortho(A: torch.Tensor, B: torch.Tensor, *,
           D: Optional[torch.Tensor] = None,
           M: Optional[LinearOperator] = None,
           mright: bool = False) -> torch.Tensor:
    # orthogonalize every column in A w.r.t. columns in B
    # D is the degeneracy map, if None, it is identity matrix
    # M is the overlap matrix (in LinearOperator)
    # mright indicates whether to operate M at the right or at the left

    # shapes:
    # A: (*BAM, na, neig)
    # B: (*BAM, na, neig)
    if D is None:
        # contracted using opt_einsum
        str1 = "...rc,...rc->...c"
        Bconj = B.conj()
        if M is None:
            return A - torch.einsum(str1, A, Bconj).unsqueeze(-2) * B
        elif mright:
            return A - torch.einsum(str1, M.mm(A), Bconj).unsqueeze(-2) * B
        else:
            return A - M.mm(torch.einsum(str1, A, Bconj).unsqueeze(-2) * B)
    else:
        BH = B.transpose(-2, -1).conj()
        if M is None:
            DBHA = D * torch.matmul(BH, A)
            return A - torch.matmul(B, DBHA)
        elif mright:
            DBHA = D * torch.matmul(BH, M.mm(A))
            return A - torch.matmul(B, DBHA)
        else:
            DBHA = D * torch.matmul(BH, A)
            return A - M.mm(torch.matmul(B, DBHA))

def custom_exacteig(A, neig, mode, M=None, **options):
    return exacteig(A, neig, mode, M)


# docstring completion
_symeig_methods = {
    "exacteig": exacteig,
    "davidson": davidson,
}
ignore_kwargs = ["M", "mparams"]
symeig.__doc__ = get_methods_docstr(symeig, _symeig_methods, ignore_kwargs)
svd.__doc__ = get_methods_docstr(svd, _symeig_methods)



__all__ = ["GetSetParamsError", "ConvergenceWarning", "MathWarning"]

class UnimplementedError(Exception):
    pass

class GetSetParamsError(Exception):
    pass

class ConvergenceWarning(Warning):
    """
    Warning to be raised if the convergence of an algorithm is not achieved
    """
    pass

class MathWarning(Warning):
    """
    Raised if there are mathematical conditions that are not satisfied
    """
    pass



import warnings
import functools
from typing import Union, Optional, Callable, Tuple, Sequence
import torch
import numpy as np
from xitorch import LinearOperator
from scipy.sparse.linalg import gmres as scipy_gmres
from xitorch._impls.optimize.root.rootsolver import broyden1
from xitorch._utils.bcast import normalize_bcast_dims, get_bcasted_dims
from xitorch._utils.exceptions import ConvergenceWarning
from xitorch._utils.types import get_np_dtype

__all__ = ["wrap_gmres", "cg", "bicgstab", "broyden1_solve", "exactsolve", "gmres"]

def wrap_gmres(A, B, E=None, M=None,
               min_eps=1e-9,
               max_niter=None,
               **unused):
    """
    Using SciPy's gmres method to solve the linear equation.

    Keyword arguments
    -----------------
    min_eps: float
        Relative tolerance for stopping conditions
    max_niter: int or None
        Maximum number of iterations. If ``None``, default to twice of the
        number of columns of ``A``.
    """
    # A: (*BA, nr, nr)
    # B: (*BB, nr, ncols)
    # E: (*BE, ncols) or None
    # M: (*BM, nr, nr) or None

    # NOTE: currently only works for batched B (1 batch dim), but unbatched A
    assert len(A.shape) == 2 and len(B.shape) == 3, "Currently only works for batched B (1 batch dim), but unbatched A"
    assert not torch.is_complex(B), "complex is not supported in gmres"

    # check the parameters
    msg = "GMRES can only do AX=B"
    assert A.shape[-2] == A.shape[-1], "GMRES can only work for square operator for now"
    assert E is None, msg
    assert M is None, msg

    nbatch, na, ncols = B.shape
    if max_niter is None:
        max_niter = 2 * na

    B = B.transpose(-1, -2)  # (nbatch, ncols, na)

    # convert the numpy/scipy
    op = A.scipy_linalg_op()
    B_np = B.detach().cpu().numpy()
    res_np = np.empty(B.shape, dtype=get_np_dtype(B.dtype))
    for i in range(nbatch):
        for j in range(ncols):
            x, info = scipy_gmres(op, B_np[i, j, :], tol=min_eps, atol=1e-12, maxiter=max_niter)
            if info > 0:
                msg = "The GMRES iteration does not converge to the desired value "\
                      "(%.3e) after %d iterations" % \
                      (min_eps, info)
                warnings.warn(ConvergenceWarning(msg))
            res_np[i, j, :] = x

    res = torch.tensor(res_np, dtype=B.dtype, device=B.device)
    res = res.transpose(-1, -2)  # (nbatch, na, ncols)
    return res

def cg(A: LinearOperator, B: torch.Tensor,
        E: Optional[torch.Tensor] = None,
        M: Optional[LinearOperator] = None,
        posdef: Optional[bool] = None,
        precond: Optional[LinearOperator] = None,
        max_niter: Optional[int] = None,
        rtol: float = 1e-6,
        atol: float = 1e-8,
        eps: float = 1e-12,
        resid_calc_every: int = 10,
        verbose: bool = False,
        **unused) -> torch.Tensor:
    r"""
    Solve the linear equations using Conjugate-Gradient (CG) method.

    Keyword arguments
    -----------------
    posdef: bool or None
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
        If None, it will be determined by power iterations.
    precond: LinearOperator or None
        LinearOperator for the preconditioning. If None, no preconditioner is
        applied.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.
    resid_calc_every: int
        Calculate the residual in its actual form instead of substitution form
        with this frequency, to avoid rounding error accummulation.
        If your linear operator has bad numerical precision, set this to be low.
        If 0, then never calculate the residual in its actual form.
    verbose: bool
        Verbosity of the algorithm.
    """
    nr = A.shape[-1]
    ncols = B.shape[-1]
    if max_niter is None:
        max_niter = int(1.5 * nr)

    # if B is all zeros, then return zeros
    batchdims = _get_batchdims(A, B, E, M)
    if torch.allclose(B, B * 0, rtol=rtol, atol=atol):
        x0 = torch.zeros((*batchdims, nr, ncols), dtype=A.dtype, device=A.device)
        return x0

    # setup the preconditioning and the matrix problem
    precond_fcn = _setup_precond(precond)
    need_hermit = True
    A_fcn, _, B2, col_swapped = _setup_linear_problem(A, B, E, M, batchdims,
                                                      posdef, need_hermit)

    # get the stopping matrix
    B_norm = B2.norm(dim=-2, keepdim=True)  # (*BB, 1, nc)
    stop_matrix = torch.max(rtol * B_norm, atol * torch.ones_like(B_norm))  # (*BB, 1, nc)

    # prepare the initial guess (it's just all zeros)
    x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr, ncols)
    xk = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

    rk = B2 - A_fcn(xk)  # (*, nr, nc)
    zk = precond_fcn(rk)  # (*, nr, nc)
    pk = zk  # (*, nr, nc)
    rkzk = _dot(rk, zk)
    converge = False
    best_resid = rk.norm(dim=-2).max().item()
    best_xk = xk
    for k in range(1, max_niter + 1):
        Apk = A_fcn(pk)
        alphak = rkzk / _safedenom(_dot(pk, Apk), eps)
        xk_1 = xk + alphak * pk

        # correct the residual calculation
        if resid_calc_every != 0 and k % resid_calc_every == 0:
            rk_1 = B2 - A_fcn(xk_1)
        else:
            rk_1 = rk - alphak * Apk  # (*, nr, nc)

        # check for the stopping condition
        resid = rk_1  # B2 - A_fcn(xk_1)
        resid_norm = resid.norm(dim=-2, keepdim=True)

        max_resid_norm = resid_norm.max().item()
        if max_resid_norm < best_resid:
            best_resid = max_resid_norm
            best_xk = xk_1

        if verbose:
            if k < 10 or k % 10 == 0:
                print("%4d: |dy|=%.3e" % (k, resid_norm))

        if torch.all(resid_norm < stop_matrix):
            converge = True
            break

        zk_1 = precond_fcn(rk_1)
        rkzk_1 = _dot(rk_1, zk_1)
        betak = rkzk_1 / _safedenom(rkzk, eps)
        pk_1 = zk_1 + betak * pk

        # move to the next index
        pk = pk_1
        zk = zk_1
        xk = xk_1
        rk = rk_1
        rkzk = rkzk_1

    xk_1 = best_xk
    if not converge:
        msg = ("Convergence is not achieved after %d iterations. "
               "Max norm of best resid: %.3e") % (max_niter, best_resid)
        warnings.warn(ConvergenceWarning(msg))
    if col_swapped:
        # x: (ncols, *, nr, 1)
        xk_1 = xk_1.transpose(0, -1).squeeze(0)  # (*, nr, ncols)
    return xk_1

def bicgstab(A: LinearOperator, B: torch.Tensor,
             E: Optional[torch.Tensor] = None,
             M: Optional[LinearOperator] = None,
             posdef: Optional[bool] = None,
             precond_l: Optional[LinearOperator] = None,
             precond_r: Optional[LinearOperator] = None,
             max_niter: Optional[int] = None,
             rtol: float = 1e-6,
             atol: float = 1e-8,
             eps: float = 1e-12,
             verbose: bool = False,
             resid_calc_every: int = 10,
             **unused) -> torch.Tensor:
    r"""
    Solve the linear equations using stabilized Biconjugate-Gradient method.

    Keyword arguments
    -----------------
    posdef: bool or None
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
        If None, it will be determined by power iterations.
    precond_l: LinearOperator or None
        LinearOperator for the left preconditioning. If None, no
        preconditioner is applied.
    precond_r: LinearOperator or None
        LinearOperator for the right preconditioning. If None, no
        preconditioner is applied.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.
    resid_calc_every: int
        Calculate the residual in its actual form instead of substitution form
        with this frequency, to avoid rounding error accummulation.
        If your linear operator has bad numerical precision, set this to be low.
        If 0, then never calculate the residual in its actual form.
    verbose: bool
        Verbosity of the algorithm.
    """
    nr, ncols = B.shape[-2:]
    if max_niter is None:
        max_niter = int(1.5 * nr)

    # if B is all zeros, then return zeros
    batchdims = _get_batchdims(A, B, E, M)
    if torch.allclose(B, B * 0, rtol=rtol, atol=atol):
        x0 = torch.zeros((*batchdims, nr, ncols), dtype=A.dtype, device=A.device)
        return x0

    # setup the preconditioning and the matrix problem
    precond_fcn_l = _setup_precond(precond_l)
    precond_fcn_r = _setup_precond(precond_r)
    need_hermit = False
    A_fcn, AT_fcn, B2, col_swapped = _setup_linear_problem(A, B, E, M, batchdims,
                                                           posdef, need_hermit)

    # get the stopping matrix
    B_norm = B2.norm(dim=-2, keepdim=True)  # (*BB, 1, nc)
    stop_matrix = torch.max(rtol * B_norm, atol * torch.ones_like(B_norm))  # (*BB, 1, nc)

    # prepare the initial guess (it's just all zeros)
    x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr, ncols)
    xk = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

    rk = B2 - A_fcn(xk)
    r0hat = rk
    rho_k = _dot(r0hat, rk)
    omega_k = torch.tensor(1.0, dtype=A.dtype, device=A.device)
    alpha: Union[float, torch.Tensor] = 1.0
    vk: Union[float, torch.Tensor] = 0.0
    pk: Union[float, torch.Tensor] = 0.0
    converge = False
    best_resid = rk.norm(dim=-2).max()
    best_xk = xk
    for k in range(1, max_niter + 1):
        rho_knew = _dot(r0hat, rk)
        omega_denom = _safedenom(omega_k, eps)
        beta = rho_knew / _safedenom(rho_k, eps) * (alpha / omega_denom)
        pk = rk + beta * (pk - omega_k * vk)
        y = precond_fcn_r(pk)
        vk = A_fcn(y)
        alpha = rho_knew / _safedenom(_dot(r0hat, vk), eps)
        h = xk + alpha * y

        s = rk - alpha * vk
        z = precond_fcn_r(s)
        t = A_fcn(z)
        Kt = precond_fcn_l(t)
        omega_k = _dot(Kt, precond_fcn_l(s)) / _safedenom(_dot(Kt, Kt), eps)
        xk = h + omega_k * z

        # correct the residual calculation regularly
        if resid_calc_every != 0 and k % resid_calc_every == 0:
            rk = B2 - A_fcn(xk)
        else:
            rk = s - omega_k * t

        # calculate the residual
        resid = rk
        resid_norm = resid.norm(dim=-2, keepdim=True)

        # save the best results
        max_resid_norm = resid_norm.max().item()
        if max_resid_norm < best_resid:
            best_resid = max_resid_norm
            best_xk = xk

        if verbose:
            if k < 10 or k % 10 == 0:
                print("%4d: |dy|=%.3e" % (k, resid_norm))

        # check for the stopping conditions
        if torch.all(resid_norm < stop_matrix):
            converge = True
            break

        rho_k = rho_knew

    xk = best_xk
    if not converge:
        msg = ("Convergence is not achieved after %d iterations. "
               "Max norm of resid: %.3e") % (max_niter, best_resid)
        warnings.warn(ConvergenceWarning(msg))
    if col_swapped:
        # x: (ncols, *, nr, 1)
        xk = xk.transpose(0, -1).squeeze(0)  # (*, nr, ncols)
    return xk

def gmres(A: LinearOperator, B: torch.Tensor,
          E: Optional[torch.Tensor] = None,
          M: Optional[LinearOperator] = None,
          posdef: Optional[bool] = None,
          max_niter: Optional[int] = None,
          rtol: float = 1e-6,
          atol: float = 1e-8,
          eps: float = 1e-12,
          **unused) -> torch.Tensor:
    r"""
    Solve the linear equations using Generalised minial residual method.

    Keyword arguments
    -----------------
    posdef: bool or None
        Indicating if the operation :math:`\mathbf{AX-MXE}` a positive
        definite for all columns and batches.
        If None, it will be determined by power iterations.
    max_niter: int or None
        Maximum number of iteration. If None, it is set to ``int(1.5 * A.shape[-1])``
    rtol: float
        Relative tolerance for stopping condition w.r.t. norm of B
    atol: float
        Absolute tolerance for stopping condition w.r.t. norm of B
    eps: float
        Substitute the absolute zero in the algorithm's denominator with this
        value to avoid nan.
    """
    converge = False

    nr, ncols = A.shape[-1], B.shape[-1]
    if max_niter is None:
        max_niter = int(nr)

    # if B is all zeros, then return zeros
    batchdims = _get_batchdims(A, B, E, M)
    if torch.allclose(B, B * 0, rtol=rtol, atol=atol):
        x0 = torch.zeros((*batchdims, nr, ncols), dtype=A.dtype, device=A.device)
        return x0

    # setup the preconditioning and the matrix problem
    need_hermit = False
    A_fcn, AT_fcn, B2, col_swapped = _setup_linear_problem(A, B, E, M, batchdims,
                                                           posdef, need_hermit)

    # get the stopping matrix
    B_norm = B2.norm(dim=-2, keepdim=True)  # (*BB, 1, nc)
    stop_matrix = torch.max(rtol * B_norm, atol * torch.ones_like(B_norm))  # (*BB, 1, nc)

    # prepare the initial guess (it's just all zeros)
    x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr, ncols)
    x0 = torch.zeros(x0shape, dtype=A.dtype, device=A.device)

    r = B2 - A_fcn(x0)  # torch.Size([*batch_dims, nr, ncols])
    best_resid = r.norm(dim=-2, keepdim=True)  # / B_norm

    best_resid = best_resid.max().item()
    best_res = x0
    q = torch.empty([max_niter] + list(r.shape), dtype=A.dtype, device=A.device)
    q[0] = r / _safedenom(r.norm(dim=-2, keepdim=True), eps)  # torch.Size([*batch_dims, nr, ncols])
    h = torch.zeros((*batchdims, ncols, max_niter + 1, max_niter), dtype=A.dtype, device=A.device)
    h = h.reshape((-1, ncols, max_niter + 1, max_niter))

    for k in range(min(nr, max_niter)):
        y = A_fcn(q[k])  # torch.Size([*batch_dims, nr, ncols])
        for j in range(k + 1):
            h[..., j, k] = _dot(q[j], y).reshape(-1, ncols)
            y = y - h[..., j, k].reshape(*batchdims, 1, ncols) * q[j]

        h[..., k + 1, k] = torch.linalg.norm(y, dim=-2)
        if torch.any(h[..., k + 1, k]) != 0 and k != max_niter - 1:
            q[k + 1] = y.reshape(-1, nr, ncols) / h[..., k + 1, k].reshape(-1, 1, ncols)
            q[k + 1] = q[k + 1].reshape(*batchdims, nr, ncols)

        b = torch.zeros((*batchdims, ncols, k + 1), dtype=A.dtype, device=A.device)
        b = b.reshape(-1, ncols, k + 1)
        b[..., 0] = torch.linalg.norm(r, dim=-2)
        rk = torch.linalg.lstsq(h[..., :k + 1, :k], b)[0]  # torch.Size([*batch_dims, max_niter])
        # Q, R = torch.linalg.qr(h[:, :k+1, :k], mode='complete')
        # result = torch.triangular_solve(torch.matmul(Q.permute(0, 2, 1), b[:, :, None])[:, :-1], R[:, :-1, :])[0]

        res = torch.empty([])
        for i in range(k):
            res = res + q[i] * rk[..., i].reshape(*batchdims, 1, ncols) + x0 if res.size() \
                else q[i] * rk[..., i].reshape(*batchdims, 1, ncols) + x0
            # res = res * B_norm

        if res.size():
            resid = B2 - A_fcn(res)
            resid_norm = resid.norm(dim=-2, keepdim=True)

            # save the best results
            max_resid_norm = resid_norm.max().item()
            if max_resid_norm < best_resid:
                best_resid = max_resid_norm
                best_res = res

            if torch.all(resid_norm < stop_matrix):
                converge = True
                break

    if not converge:
        msg = ("Convergence is not achieved after %d iterations. "
               "Max norm of resid: %.3e") % (max_niter, best_resid)
        warnings.warn(ConvergenceWarning(msg))

    res = best_res
    return res


############ cg and bicgstab helpers ############
def _safedenom(r: torch.Tensor, eps: float) -> torch.Tensor:
    r[r == 0] = eps
    return r

def _dot(r: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    # r: (*BR, nr, nc)
    # z: (*BR, nr, nc)
    # return: (*BR, 1, nc)
    return torch.einsum("...rc,...rc->...c", r.conj(), z).unsqueeze(-2)

############ rootfinder-based ############
@functools.wraps(broyden1)
def broyden1_solve(A, B, E=None, M=None, **options):
    return _rootfinder_solve("broyden1", A, B, E, M, **options)

def _rootfinder_solve(alg, A, B, E=None, M=None, **options):
    # using rootfinder algorithm
    nr = A.shape[-1]
    ncols = B.shape[-1]

    # set up the function for the rootfinding
    def fcn_rootfinder(xi):
        # xi: (*BX, nr*ncols)
        x = xi.reshape(*xi.shape[:-1], nr, ncols)  # (*BX, nr, ncols)
        y = A.mm(x) - B  # (*BX, nr, ncols)
        if E is not None:
            MX = M.mm(x) if M is not None else x
            MXE = MX * E.unsqueeze(-2)
            y = y - MXE  # (*BX, nr, ncols)
        y = y.reshape(*xi.shape[:-1], -1)  # (*BX, nr*ncols)
        return y

    # setup the initial guess (the batch dimension must be the largest)
    batchdims = _get_batchdims(A, B, E, M)
    x0 = torch.zeros((*batchdims, nr * ncols), dtype=A.dtype, device=A.device)

    if alg == "broyden1":
        x = broyden1(fcn_rootfinder, x0, **options)
    else:
        raise RuntimeError("Unknown method %s" % alg)
    x = x.reshape(*x.shape[:-1], nr, ncols)
    return x

############ exact solve ############
def exactsolve(A: LinearOperator, B: torch.Tensor,
               E: Union[torch.Tensor, None],
               M: Union[LinearOperator, None]):
    """
    Solve the linear equation by contructing the full matrix of LinearOperators.

    Warnings
    --------
    * As this method construct the linear operators explicitly, it might requires
      a large memory.
    """
    # A: (*BA, na, na)
    # B: (*BB, na, ncols)
    # E: (*BE, ncols)
    # M: (*BM, na, na)
    if E is None:
        Amatrix = A.fullmatrix()  # (*BA, na, na)
        x = torch.linalg.solve(Amatrix, B)  # (*BAB, na, ncols)
    elif M is None:
        Amatrix = A.fullmatrix()
        x = _solve_ABE(Amatrix, B, E)
    else:
        Mmatrix = M.fullmatrix()  # (*BM, na, na)
        L = torch.linalg.cholesky(Mmatrix)  # (*BM, na, na)
        Linv = torch.inverse(L)  # (*BM, na, na)
        LinvT = Linv.transpose(-2, -1).conj()  # (*BM, na, na)
        A2 = torch.matmul(Linv, A.mm(LinvT))  # (*BAM, na, na)
        B2 = torch.matmul(Linv, B)  # (*BBM, na, ncols)

        X2 = _solve_ABE(A2, B2, E)  # (*BABEM, na, ncols)
        x = torch.matmul(LinvT, X2)  # (*BABEM, na, ncols)
    return x

def _solve_ABE(A: torch.Tensor, B: torch.Tensor, E: torch.Tensor):
    # A: (*BA, na, na) matrix
    # B: (*BB, na, ncols) matrix
    # E: (*BE, ncols) matrix
    na = A.shape[-1]
    BA, BB, BE = normalize_bcast_dims(A.shape[:-2], B.shape[:-2], E.shape[:-1])
    E = E.reshape(1, *BE, E.shape[-1]).transpose(0, -1)  # (ncols, *BE, 1)
    B = B.reshape(1, *BB, *B.shape[-2:]).transpose(0, -1)  # (ncols, *BB, na, 1)

    # NOTE: The line below is very inefficient for large na and ncols
    AE = A - torch.diag_embed(E.repeat_interleave(repeats=na, dim=-1), dim1=-2, dim2=-1)  # (ncols, *BAE, na, na)
    r = torch.linalg.solve(AE, B)  # (ncols, *BAEM, na, 1)
    r = r.transpose(0, -1).squeeze(0)  # (*BAEM, na, ncols)
    return r

############ general helpers ############
def _get_batchdims(A: LinearOperator, B: torch.Tensor,
                   E: Union[torch.Tensor, None],
                   M: Union[LinearOperator, None]):

    batchdims = [A.shape[:-2], B.shape[:-2]]
    if E is not None:
        batchdims.append(E.shape[:-1])
        if M is not None:
            batchdims.append(M.shape[:-2])
    return get_bcasted_dims(*batchdims)

def _setup_precond(precond: Optional[LinearOperator]) -> Callable[[torch.Tensor], torch.Tensor]:
    if isinstance(precond, LinearOperator):
        precond_fcn = lambda x: precond.mm(x)
    elif precond is None:
        precond_fcn = lambda x: x
    else:
        raise TypeError("precond can only be LinearOperator or None")
    return precond_fcn

def _setup_linear_problem(A: LinearOperator, B: torch.Tensor,
                          E: Optional[torch.Tensor], M: Optional[LinearOperator],
                          batchdims: Sequence[int],
                          posdef: Optional[bool],
                          need_hermit: bool) -> \
        Tuple[Callable[[torch.Tensor], torch.Tensor],
              Callable[[torch.Tensor], torch.Tensor],
              torch.Tensor, bool]:

    # get the linear operator (including the MXE part)
    if E is None:
        A_fcn = lambda x: A.mm(x)
        AT_fcn = lambda x: A.rmm(x)
        B_new = B
        col_swapped = False
    else:
        # A: (*BA, nr, nr) linop
        # B: (*BB, nr, ncols)
        # E: (*BE, ncols)
        # M: (*BM, nr, nr) linop
        if M is None:
            BAs, BBs, BEs = normalize_bcast_dims(A.shape[:-2], B.shape[:-2], E.shape[:-1])
        else:
            BAs, BBs, BEs, BMs = normalize_bcast_dims(A.shape[:-2], B.shape[:-2],
                                                      E.shape[:-1], M.shape[:-2])
        E = E.reshape(*BEs, *E.shape[-1:])
        E_new = E.unsqueeze(0).transpose(-1, 0).unsqueeze(-1)  # (ncols, *BEs, 1, 1)
        B = B.reshape(*BBs, *B.shape[-2:])  # (*BBs, nr, ncols)
        B_new = B.unsqueeze(0).transpose(-1, 0)  # (ncols, *BBs, nr, 1)

        def A_fcn(x):
            # x: (ncols, *BX, nr, 1)
            Ax = A.mm(x)  # (ncols, *BAX, nr, 1)
            Mx = M.mm(x) if M is not None else x  # (ncols, *BMX, nr, 1)
            MxE = Mx * E_new  # (ncols, *BMXE, nr, 1)
            return Ax - MxE

        def AT_fcn(x):
            # x: (ncols, *BX, nr, 1)
            ATx = A.rmm(x)
            MTx = M.rmm(x) if M is not None else x
            MTxE = MTx * E_new
            return ATx - MTxE

        col_swapped = True

    # estimate if it's posdef with power iteration
    if need_hermit:
        is_hermit = A.is_hermitian and (M is None or M.is_hermitian)
        if not is_hermit:
            # set posdef to False to make the operator becomes AT * A so it is
            # hermitian
            posdef = False

    # TODO: the posdef check by largest eival only works for Hermitian/symmetric
    # matrix, but it doesn't always work for non-symmetric matrix.
    # In non-symmetric case, one need to do Cholesky LDL decomposition
    if posdef is None:
        nr, ncols = B.shape[-2:]
        x0shape = (ncols, *batchdims, nr, 1) if col_swapped else (*batchdims, nr, ncols)
        x0 = torch.randn(x0shape, dtype=A.dtype, device=A.device)
        x0 = x0 / x0.norm(dim=-2, keepdim=True)
        largest_eival = _get_largest_eival(A_fcn, x0)  # (*, 1, nc)
        negeival = largest_eival <= 0

        # if the largest eigenvalue is negative, then it's not posdef
        if torch.all(negeival):
            posdef = False

        # otherwise, calculate the lowest eigenvalue to check if it's positive
        else:
            offset = torch.clamp(largest_eival, min=0.0)
            A_fcn2 = lambda x: A_fcn(x) - offset * x
            mostneg_eival = _get_largest_eival(A_fcn2, x0)  # (*, 1, nc)
            posdef = bool(torch.all(torch.logical_or(-mostneg_eival <= offset, negeival)).item())

    # get the linear operation if it is not a posdef (A -> AT.A)
    if posdef:
        return A_fcn, AT_fcn, B_new, col_swapped
    else:
        def A_new_fcn(x):
            return AT_fcn(A_fcn(x))
        B2 = AT_fcn(B_new)
        return A_new_fcn, A_new_fcn, B2, col_swapped

def _get_largest_eival(Afcn, x):
    niter = 10
    rtol = 1e-3
    atol = 1e-6
    xnorm_prev = None
    for i in range(niter):
        x = Afcn(x)  # (*, nr, nc)
        xnorm = x.norm(dim=-2, keepdim=True)  # (*, 1, nc)

        # check if xnorm is converging
        if i > 0:
            dnorm = torch.abs(xnorm_prev - xnorm)
            if torch.all(dnorm <= rtol * xnorm + atol):
                break

        xnorm_prev = xnorm
        if i < niter - 1:
            x = x / xnorm
    return xnorm



import torch
import copy
from typing import Callable, Union, Mapping, Any, Sequence, Dict
from xitorch._utils.assertfuncs import assert_fcn_params, assert_runtime
from xitorch._core.pure_function import get_pure_function, make_sibling
from xitorch._impls.integrate.ivp.explicit_rk import rk4_ivp, rk38_ivp, fwd_euler_ivp
from xitorch._impls.integrate.ivp.adaptive_rk import rk23_adaptive, rk45_adaptive
from xitorch._utils.misc import set_default_option, TensorNonTensorSeparator, \
    TensorPacker, get_method
from xitorch._utils.tensor import convert_none_grads_to_zeros
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch.debug.modes import is_debug_enabled

__all__ = ["solve_ivp"]

def solve_ivp(fcn: Union[Callable[..., torch.Tensor], Callable[..., Sequence[torch.Tensor]]],
              ts: torch.Tensor,
              y0: torch.Tensor,
              params: Sequence[Any] = [],
              bck_options: Mapping[str, Any] = {},
              method: Union[str, Callable, None] = None,
              **fwd_options) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
    r"""
    Solve the initial value problem (IVP) or also commonly known as ordinary
    differential equations (ODE), where given the initial value :math:`\mathbf{y_0}`,
    it then solves

    .. math::

        \mathbf{y}(t) = \mathbf{y_0} + \int_{t_0}^{t} \mathbf{f}(t', \mathbf{y}, \theta)\ \mathrm{d}t'

    Although the original ``solve_ivp`` does not accept batched ``ts``, it can
    be batched using functorch's ``vmap`` (only for explicit solver, though,
    e.g. ``rk38``, ``rk4``, and ``euler``). Adaptive steps cannot be vmapped
    at the moment.

    Arguments
    ---------
    fcn: callable
        The function that represents dy/dt. The function takes an input of a
        single time ``t`` and tensor ``y`` with shape ``(*ny)`` and
        produce :math:`\mathrm{d}\mathbf{y}/\mathrm{d}t` with shape ``(*ny)``.
        The output of the function must be a tensor with shape ``(*ny)`` or
        a list of tensors.
    ts: torch.tensor
        The time points where the value of `y` will be returned.
        It must be monotonically increasing or decreasing.
        It is a tensor with shape ``(nt,)``.
    y0: torch.tensor
        The initial value of ``y``, i.e. ``y(t[0]) == y0``.
        It is a tensor with shape ``(*ny)`` or a list of tensors.
    params: list
        Sequence of other parameters required in the function.
    bck_options: dict
        Options for the backward solve_ivp method. If not specified, it will
        take the same options as fwd_options.
    method: str or callable or None
        Initial value problem solver. If None, it will choose ``"rk45"``.
    **fwd_options
        Method-specific option (see method section below).

    Returns
    -------
    torch.tensor or a list of tensors
        The values of ``y`` for each time step in ``ts``.
        It is a tensor with shape ``(nt,*ny)`` or a list of tensors
    """
    if is_debug_enabled():
        assert_fcn_params(fcn, (ts[0], y0, *params))
    assert_runtime(len(ts.shape) == 1, "Argument ts must be a 1D tensor")

    if method is None:  # set the default method
        method = "rk45"
    fwd_options["method"] = method

    is_y0_list = isinstance(y0, (list, tuple))
    pfcn = get_pure_function(fcn)
    if is_y0_list:
        nt = len(ts)
        roller = TensorPacker(y0)

        @make_sibling(pfcn)
        def pfcn2(t, ytensor, *params):
            ylist = roller.pack(ytensor)
            res_list = pfcn(t, ylist, *params)
            if not isinstance(res_list, (list, tuple)):
                raise RuntimeError("The y0 and output of fcn must both be tuple or a tensor")
            res = roller.flatten(res_list)
            return res

        y0 = roller.flatten(y0)
        res = _SolveIVP.apply(pfcn2, ts, fwd_options, bck_options, len(params), y0, *params, *pfcn.objparams())
        return roller.pack(res)
    else:
        return _SolveIVP.apply(pfcn, ts, fwd_options, bck_options, len(params), y0, *params, *pfcn.objparams())

class _SolveIVP(torch.autograd.Function):
    @staticmethod
    def forward(ctx, pfcn, ts, fwd_options, bck_options, nparams, y0, *allparams):
        config = fwd_options
        ctx.bck_config = set_default_option(config, bck_options)

        params = allparams[:nparams]
        objparams = allparams[nparams:]

        method = config.pop("method")
        methods = {
            "rk4": rk4_ivp,
            "rk38": rk38_ivp,
            "rk23": rk23_adaptive,
            "rk45": rk45_adaptive,
            "euler": fwd_euler_ivp,
        }
        solver = get_method("solve_ivp", methods, method)
        yt = solver(pfcn, ts, y0, params, **config)

        # save the parameters for backward
        ctx.param_sep = TensorNonTensorSeparator(allparams, varonly=True)
        tensor_params = ctx.param_sep.get_tensor_params()
        ctx.save_for_backward(ts, y0, *tensor_params)
        ctx.pfcn = pfcn
        ctx.nparams = nparams
        ctx.yt = yt
        ctx.ts_requires_grad = ts.requires_grad

        return yt

    @staticmethod
    def backward(ctx, grad_yt):
        # grad_yt: (nt, *ny)
        nparams = ctx.nparams
        pfcn = ctx.pfcn
        param_sep = ctx.param_sep
        yt = ctx.yt
        ts_requires_grad = ctx.ts_requires_grad

        # restore the parameters
        saved_tensors = ctx.saved_tensors
        ts = saved_tensors[0]
        y0 = saved_tensors[1]
        tensor_params = list(saved_tensors[2:])
        allparams = param_sep.reconstruct_params(tensor_params)
        ntensor_params = len(tensor_params)
        params = allparams[:nparams]
        objparams = allparams[nparams:]

        grad_enabled = torch.is_grad_enabled()
        # custom function to evaluate the input `pfcn` based on whether we want
        # to connect the graph or not

        def pfunc2(t, y, tensor_params):
            if not grad_enabled:
                # if graph is not constructed, then use the default tensor_params
                ycopy = y.detach().requires_grad_()  # [yi.detach().requires_grad_() for yi in y]
                tcopy = t.detach().requires_grad_()
                f = pfcn(tcopy, ycopy, *params)
                return f, tcopy, ycopy, tensor_params
            else:
                # if graph is constructed, then use the clone of the tensor params
                # so that infinite loop of backward can be avoided
                tensor_params_copy = [p.clone().requires_grad_() for p in tensor_params]
                ycopy = y.clone().requires_grad_()
                tcopy = t.clone().requires_grad_()
                allparams_copy = param_sep.reconstruct_params(tensor_params_copy)
                params_copy = allparams_copy[:nparams]
                objparams_copy = allparams_copy[nparams:]
                with pfcn.useobjparams(objparams_copy):
                    f = pfcn(tcopy, ycopy, *params_copy)
                return f, tcopy, ycopy, tensor_params_copy

        # slices and indices definitions on the augmented states
        y_index = 0
        dLdy_index = 1
        dLdt_index = 2
        dLdt_slice = slice(dLdt_index, dLdt_index + 1, None)  # [2:3]
        dLdp_slice = slice(-ntensor_params, None, None) if ntensor_params > 0 else slice(0,
                                                                                         0, None)  # [-ntensor_params:]
        state_size = 3 + ntensor_params
        states = [None for _ in range(state_size)]

        def new_pfunc(t, states, *tensor_params):
            # t: single-element
            y = states[y_index]
            dLdy = -states[dLdy_index]
            with torch.enable_grad():
                f, t2, y2, tensor_params2 = pfunc2(t, y, tensor_params)
            allgradinputs = ([y2] + [t2] + list(tensor_params2))
            allgrads = torch.autograd.grad(f,
                                           inputs=allgradinputs,
                                           grad_outputs=dLdy,
                                           retain_graph=True,
                                           allow_unused=True,
                                           create_graph=torch.is_grad_enabled())  # list of (*ny)
            allgrads = convert_none_grads_to_zeros(allgrads, allgradinputs)
            outs = (
                f,  # dydt
                *allgrads,
            )
            return outs

        ts_flip = ts.flip(0)
        t_flip_idx = -1
        states[y_index] = yt[t_flip_idx]
        states[dLdy_index] = grad_yt[t_flip_idx]
        states[dLdt_index] = torch.zeros_like(ts[0])
        states[dLdp_slice] = [torch.zeros_like(tp) for tp in tensor_params]
        grad_ts = [None for _ in range(len(ts))] if ts_requires_grad else None

        # define a new function for the augmented dynamics
        bkw_roller = TensorPacker(states)

        @make_sibling(new_pfunc)
        def pfcn_back(t, ytensor, *params):
            ylist = bkw_roller.pack(ytensor)
            res_list = new_pfunc(t, ylist, *params)
            res = bkw_roller.flatten(res_list)
            return res

        for i in range(len(ts_flip) - 1):
            if ts_requires_grad:
                feval = pfunc2(ts_flip[i], states[y_index], tensor_params)[0]
                dLdt1 = torch.dot(feval.reshape(-1), grad_yt[t_flip_idx].reshape(-1))
                states[dLdt_index] -= dLdt1
                grad_ts[t_flip_idx] = dLdt1.reshape(-1)

            t_flip_idx -= 1
            states_flatten = bkw_roller.flatten(states)
            fwd_config = copy.copy(ctx.bck_config)
            bck_config = copy.copy(ctx.bck_config)
            outs_flatten = _SolveIVP.apply(
                pfcn_back, ts_flip[i:i + 2], fwd_config, bck_config, len(tensor_params),
                states_flatten, *tensor_params)
            outs = bkw_roller.pack(outs_flatten)

            # only take the output for the earliest time
            states = [out[-1] for out in outs]
            states[y_index] = yt[t_flip_idx]
            # gyt is the contribution from the input grad_y
            # gy0 is the propagated gradients from the later time step
            states[dLdy_index] = grad_yt[t_flip_idx] + states[dLdy_index]

        if ts_requires_grad:
            grad_ts[0] = states[dLdt_index].reshape(-1)

        grad_y0 = states[dLdy_index]  # dL/dy0, (*ny)
        if ts_requires_grad:
            grad_ts = torch.cat(grad_ts).reshape(*ts.shape)
        grad_tensor_params = states[dLdp_slice]
        grad_ntensor_params = [None for _ in range(len(allparams) - ntensor_params)]
        grad_params = param_sep.reconstruct_params(grad_tensor_params, grad_ntensor_params)
        return (None, grad_ts, None, None, None, grad_y0, *grad_params)


# docstring completion
ivp_methods: Dict[str, Callable] = {
    "rk45": rk45_adaptive,
    "rk23": rk23_adaptive,
    "rk4": rk4_ivp,
    "rk38": rk38_ivp,
    "euler": fwd_euler_ivp,
}
solve_ivp.__doc__ = get_methods_docstr(solve_ivp, ivp_methods)



import torch
import numpy as np

################### metropolis hastings ###################
def mh(logpfcn, x0, pparams, nsamples=10000, nburnout=5000, step_size=1.0, **unused):
    """
    Perform Metropolis-Hasting steps to collect samples

    Keyword arguments
    -----------------
    nsamples: int
        The number of samples to be collected
    nburnout: int
        The number of initial steps to be performed before collecting samples
    step_size: float
        The size of the steps to be taken
    """
    x, dtype, device = _mh_sample(logpfcn, x0, pparams, nburnout, step_size, False)
    samples = _mh_sample(logpfcn, x, pparams, nsamples, step_size, True)
    weights = torch.zeros((samples.shape[0],), dtype=dtype, device=device) + (1. / samples.shape[0])
    return samples, weights

def _mh_sample(logpfcn, x0, pparams, nsamples, step_size, collect_samples):
    x = x0
    logpx = logpfcn(x0, *pparams)
    dtype = logpx.dtype
    device = logpx.device
    log_rand = torch.log(torch.rand((nsamples,), dtype=dtype, device=device))
    if collect_samples:
        samples = torch.empty((nsamples, *x0.shape), dtype=x.dtype, device=x.device)

    for i in range(nsamples):
        xnext = x + step_size * torch.randn_like(x)
        logpnext = logpfcn(xnext, *pparams)
        logpratio = logpnext - logpx

        # decide if we should accept the next point
        if logpratio > 0:
            accept = True
        else:
            accept = log_rand[i] < logpratio

        # if accept, move the x into the new points
        if accept:
            logpx = logpnext
            x = xnext
        if collect_samples:
            samples[i] = x

    # return the samples if collect_samples, otherwise just return the last x
    if collect_samples:
        return samples
    else:
        return x, dtype, device

def mhcustom(logpfcn, x0, pparams, nsamples=10000, nburnout=5000, custom_step=None, **unused):
    """
    Perform Metropolis sampling using custom_step

    Keyword arguments
    -----------------
    nsamples: int
        The number of samples to be collected
    nburnout: int
        The number of initial steps to be performed before collecting samples
    custom_step: callable or None
        Callable with call signature ``custom_step(x, *pparams)`` to produce the
        next samples (already decided whether to accept or not).
        This argument is **required**. If ``None``, it will raise an error
    """
    if custom_step is None:
        raise RuntimeError("custom_step must be specified for mhcustom method")
    if not hasattr(custom_step, "__call__"):
        raise RuntimeError("custom_step option for mhcustom must be callable")

    x, dtype, device = _mhcustom_sample(logpfcn, x0, pparams, nburnout, custom_step, False)
    xsamples = _mhcustom_sample(logpfcn, x0, pparams, nburnout, custom_step, True)
    wsamples = torch.zeros((xsamples.shape[0],), dtype=dtype, device=device) + (1. / xsamples.shape[0])
    return xsamples, wsamples

def _mhcustom_sample(logpfcn, x0, pparams, nsamples, custom_step, collect_samples):
    x = x0
    logpx = logpfcn(x0, *pparams)
    dtype = logpx.dtype
    device = logpx.device
    if collect_samples:
        samples = torch.empty((nsamples, *x0.shape), dtype=x.dtype, device=x.device)
        samples[0] = x

    for i in range(1, nsamples):
        x = custom_step(x, *pparams)
        if collect_samples:
            samples[i] = x
    if collect_samples:
        return samples
    else:
        return x, dtype, device

################### dummy sampler just for 1D ###################
def dummy1d(logpfcn, x0, pparams, nsamples=100, lb=-np.inf, ub=np.inf, **unused):
    dtype = x0.dtype
    device = x0.device

    # convert the bound to finite range
    ub = torch.tensor(ub, dtype=dtype, device=device)
    lb = torch.tensor(lb, dtype=dtype, device=device)
    tu = torch.atan(ub)
    tl = torch.atan(lb)

    assert torch.numel(x0) == 1, "This dummy operation can only be done in 1D space"
    tlg, wlg = np.polynomial.legendre.leggauss(nsamples)
    tlg = torch.tensor(tlg, dtype=dtype, device=device)
    wlg = torch.tensor(wlg, dtype=dtype, device=device)
    wlg *= 0.5 * (tu - tl)
    tsamples = tlg * (0.5 * (tu - tl)) + (0.5 * (tu + tl))  # (n, *nx)
    xsamples = torch.tan(tsamples)
    wt = torch.cos(tsamples)**(-2.)
    wp = torch.empty_like(wt)
    for i in range(nsamples):
        wp[i] = torch.exp(logpfcn(xsamples[i], *pparams))

    wsamples = wt * wlg * wp
    wsamples = wsamples / wsamples.sum()
    return xsamples, wsamples



from typing import Optional
import torch
import warnings
from abc import abstractmethod
from xitorch._impls.interpolate.base_interp import BaseInterp
from xitorch._impls.interpolate.extrap_utils import get_extrap_pos, get_extrap_val
from xitorch._utils.bcast import match_dim

class BaseInterp1D(BaseInterp):
    def __init__(self, x: torch.Tensor, y: Optional[torch.Tensor] = None, extrap: Optional[str] = None,
                 **unused):
        # x: (*BX, nr)
        # y: (*BY, nr), BX and BY are broadcastable
        self._y_is_given = y is not None
        self._extrap = extrap
        self._xmin = torch.min(x, dim=-1, keepdim=True)[0]
        self._xmax = torch.max(x, dim=-1, keepdim=True)[0]
        self._is_periodic_required = False
        self._y = y

    def set_periodic_required(self, val):
        self._is_periodic_required = val

    def is_periodic_required(self):
        return self._is_periodic_required

    def __call__(self, xq: torch.Tensor, y: Optional[torch.Tensor] = None) -> torch.Tensor:
        # xq: (*BX, nrq)
        # y: (*BY, nr)
        if self._y_is_given and y is not None:
            msg = "y has been supplied when initiating this instance. This value of y will be ignored"
            # stacklevel=3 because this __call__ will be called by a wrapper's __call__
            warnings.warn(msg, stacklevel=3)

        extrap = self._extrap
        if self._y_is_given:
            y = self._y
        elif y is None:
            raise RuntimeError("y must be given")
        elif self.is_periodic_required():
            check_periodic_value(y)
        assert y is not None

        xqinterp_mask = torch.logical_and(xq >= self._xmin, xq <= self._xmax)  # (*BX, nrq)
        xqextrap_mask = ~xqinterp_mask
        allinterp = torch.all(xqinterp_mask)

        if not allinterp and xqextrap_mask.ndim > 1:
            raise NotImplementedError("Batched interpolation + extrapolation has not been implemented yet")

        if allinterp:
            return self._interp(xq, y=y)
        elif extrap == "mirror" or extrap == "periodic" or extrap == "bound":
            # extrapolation by mapping it to the interpolated region
            xq2 = xq.clone()
            xq2[xqextrap_mask] = get_extrap_pos(xq[xqextrap_mask], extrap, self._xmin, self._xmax)
            return self._interp(xq2, y=y)
        else:
            # interpolation
            yqinterp = self._interp(xq[xqinterp_mask], y=y)  # (*BY, nrq)
            yqextrap = get_extrap_val(xq[xqextrap_mask], y, extrap)

            yq = torch.empty((*y.shape[:-1], xq.shape[-1]), dtype=y.dtype, device=y.device)  # (*BY, nrq)
            yq[..., xqinterp_mask] = yqinterp
            yq[..., xqextrap_mask] = yqextrap
            return yq

    @abstractmethod
    def _interp(self, xq: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        pass

class CubicSpline1D(BaseInterp1D):
    def __init__(self, x: torch.Tensor, y: Optional[torch.Tensor] = None,
                 bc_type: Optional[str] = None, extrap: Optional[str] = None, **unused):
        # x: (*BX, nr)
        # y: (*BY, nr), BX and BY are broadcastable

        # get the default extrapolation method and boundary condition
        if bc_type is None:
            bc_type = "not-a-knot"
        extrap = check_and_get_extrap(extrap, bc_type)
        super(CubicSpline1D, self).__init__(x, y, extrap=extrap)

        self.x = x
        # if x.ndim != 1:
        #     raise RuntimeError("The input x must be a 1D tensor")

        bc_types = ["natural", "clamped", "not-a-knot", "periodic"]
        if bc_type not in bc_types:
            raise RuntimeError("Unimplemented %s bc_type. Available options: %s" % (bc_type, bc_types))
        self.bc_type = bc_type
        self.set_periodic_required(extrap == "periodic")  # or self.bc_type == "periodic"

        # precompute the inverse of spline matrix
        self.spline_mat_inv = _get_spline_mat_inv(x, bc_type)  # (*BX, nr, nr)
        self.y_is_given = y is not None
        if self.y_is_given:
            if self.is_periodic_required():
                check_periodic_value(y)
            self.y = y
            assert y is not None
            self.ks = torch.matmul(self.spline_mat_inv, y.unsqueeze(-1)).squeeze(-1)

    def _interp(self, xq, y):
        # https://en.wikipedia.org/wiki/Spline_interpolation#Algorithm_to_find_the_interpolating_cubic_spline
        # get the k-vector (i.e. the gradient at every points)
        if self.y_is_given:
            ks = self.ks
        else:
            ks = torch.matmul(self.spline_mat_inv, y.unsqueeze(-1)).squeeze(-1)  # (*BY, nr)

        x, xq = match_dim(self.x, xq, contiguous=True)  # (*BX, nr)

        # find the index location of xq
        nr = x.shape[-1]
        # detaching due to PyTorch's issue #42328
        idxr = torch.searchsorted(x.detach(), xq.detach(), right=False)  # (*BX, nrq)
        idxr = torch.clamp(idxr, 1, nr - 1)
        idxl = idxr - 1  # (*BX, nrq) from (0 to nr-2)

        if torch.numel(xq) > torch.numel(x):
            # get the variables needed
            yl = y[..., :-1]  # (*BY, nr-1)
            xl = x[..., :-1]  # (nr-1)
            dy = y[..., 1:] - yl  # (*BY, nr-1)
            dx = x[..., 1:] - xl  # (nr-1)
            a = ks[..., :-1] * dx - dy  # (*BY, nr-1)
            b = -ks[..., 1:] * dx + dy  # (*BY, nr-1)

            # calculate the coefficients for the t-polynomial
            p0 = yl  # (*BY, nr-1)
            p1 = (dy + a)  # (*BY, nr-1)
            p2 = (b - 2 * a)  # (*BY, nr-1)
            p3 = a - b  # (*BY, nr-1)

            t = (xq - torch.gather(xl, -1, idxl)) / torch.gather(dx, -1, idxl)  # (*BX, nrq)
            # yq = p0[:,idxl] + t * (p1[:,idxl] + t * (p2[:,idxl] + t * p3[:,idxl])) # (nbatch, nrq)
            p0, p1, p2, p3, idxl = match_dim(p0, p1, p2, p3, idxl)
            yq = torch.gather(p3, dim=-1, index=idxl) * t
            yq += torch.gather(p2, dim=-1, index=idxl)
            yq *= t
            yq += torch.gather(p1, dim=-1, index=idxl)
            yq *= t
            yq += torch.gather(p0, dim=-1, index=idxl)
            return yq

        else:
            x, y, ks, idxl, idxr = match_dim(x, y, ks, idxl, idxr)
            xl = torch.gather(x, dim=-1, index=idxl)
            xr = torch.gather(x, dim=-1, index=idxr)
            yl = torch.gather(y, dim=-1, index=idxl).contiguous()
            yr = torch.gather(y, dim=-1, index=idxr).contiguous()
            kl = torch.gather(ks, dim=-1, index=idxl).contiguous()
            kr = torch.gather(ks, dim=-1, index=idxr).contiguous()

            dxrl = xr - xl  # (nrq,)
            # dyrl = yr - yl  # (nbatch, nrq)

            # calculate the coefficients of the large matrices
            t = (xq - xl) / dxrl  # (nrq,)
            tinv = 1 - t  # nrq
            tta = t * tinv * tinv
            ttb = t * tinv * t
            tyl = tinv + tta - ttb
            tyr = t - tta + ttb
            tkl = tta * dxrl
            tkr = -ttb * dxrl

            yq = yl * tyl + yr * tyr + kl * tkl + kr * tkr
            return yq

    def getparamnames(self):
        if self.y_is_given:
            res = ["x", "y", "ks"]
        else:
            res = ["spline_mat_inv", "x"]
        return res

class LinearInterp1D(BaseInterp1D):
    def __init__(self, x, y=None, extrap=None, **unused):
        super(LinearInterp1D, self).__init__(x, y, extrap=extrap)
        self.x = x
        self.y_is_given = y is not None
        self.y = y

    def _interp(self, xq: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        if self.y_is_given:
            y = self.y

        x, xq = match_dim(self.x, xq, contiguous=True)

        nr = x.shape[-1]
        idxr = torch.searchsorted(x.detach(), xq.detach(), right=False)  # (nrq)
        idxr = torch.clamp(idxr, 1, nr - 1)
        idxl = idxr - 1  # (nrq) from (0 to nr-2)

        if torch.numel(xq) > torch.numel(x):
            x, y, idxl = match_dim(x, y, idxl)
            yl = y[..., :-1]  # (..., nr-1)
            xl = x[..., :-1]  # (..., nr-1)
            dy = y[..., 1:] - yl  # (..., nr-1)
            dx = x[..., 1:] - xl  # (..., nr-1)
            t = (xq - torch.gather(xl, -1, idxl)) / torch.gather(dx, -1, idxl)  # (..., nrq)
            yq = torch.gather(dy, dim=-1, index=idxl) * t
            yq += torch.gather(yl, dim=-1, index=idxl)
            return yq
        else:
            x, y, idxl, idxr = match_dim(x, y, idxl, idxr)
            xl = torch.gather(x, -1, idxl)
            xr = torch.gather(x, -1, idxr)
            yl = torch.gather(y, dim=-1, index=idxl).contiguous()
            yr = torch.gather(y, dim=-1, index=idxr).contiguous()

            dxrl = xr - xl  # (..., nrq)
            dyrl = yr - yl  # (..., nrq)
            t = (xq - xl) / dxrl  # (..., nrq)
            yq = yl + dyrl * t
            return yq

    def getparamnames(self):
        if self.y_is_given:
            res = ["x", "y"]
        else:
            res = ["x"]
        return res

##### docstrings #####
extrap_docstr = """
    extrap: int, float, 1-element torch.Tensor, str, or None
        Extrapolation option:

        * ``int``, ``float``, or 1-element ``torch.Tensor``: it will pad the extrapolated
          values with the specified values
        * ``"mirror"``: the extrapolation values are mirrored
        * ``"periodic"``: periodic boundary condition. ``y[...,0] == y[...,-1]`` must
          be fulfilled for this condition.
        * ``"bound"``: fill in the extrapolated values with the left or right bound
          values.
        * ``"nan"``: fill the extrapolated values with nan
        * callable: apply this extrapolation function with the extrapolated
          positions and use the output as the values
        * ``None``: choose the extrapolation based on the ``bc_type``. These are the
          pairs:

          * ``"clamped"``: ``"mirror"``
          * other: ``"nan"``

        Default: ``None``"""

CubicSpline1D.__doc__ = """
    Perform 1D cubic spline interpolation for non-uniform ``x`` [1]_ [2]_.

    Keyword arguments
    -----------------
    bc_type: str or None
        Boundary condition:

        * ``"not-a-knot"``: The first and second segments are the same polynomial
        * ``"natural"``: 2nd grad at the boundaries are 0
        * ``"clamped"``: 1st grad at the boundaries are 0
        * ``"periodic"``: periodic boundary condition (`new in version 0.2`)

        If ``None``, it will choose ``"not-a-knot"``
    """ + extrap_docstr + """

    References
    ----------
    .. [1] SplineInterpolation on Wikipedia,
           https://en.wikipedia.org/wiki/Spline_interpolation#Algorithm_to_find_the_interpolating_cubic_spline)
    .. [2] Carl de Boor, "A Practical Guide to Splines", Springer-Verlag, 1978.
"""

LinearInterp1D.__doc__ = """
    Perform 1D linear interpolation for non-uniform ``x``.

    Keyword arguments
    -----------------""" + extrap_docstr + """
"""

def check_and_get_extrap(extrap, bc_type):
    if extrap is None:
        try:
            return {
                "clamped": "mirror",
                "periodic": "periodic",
            }[bc_type]
        except KeyError:
            return "nan"
    return extrap

def check_periodic_value(y):
    if not torch.allclose(y[..., 0], y[..., -1]):
        raise RuntimeError("The value of y must be periodic to have periodic bc_type or extrap")

# @torch.jit.script
def _get_spline_mat_inv(x: torch.Tensor, bc_type: str):
    """
    Returns the inverse of spline matrix where the gradient can be obtained just
    by

    >>> spline_mat_inv = _get_spline_mat_inv(x, transpose=True)
    >>> ks = torch.matmul(y, spline_mat_inv)

    where `y` is a tensor of (nbatch, nr) and `spline_mat_inv` is the output of
    this function with shape (nr, nr)

    Arguments
    ---------
    x: torch.Tensor with shape (*BX, nr)
        The x-position of the data
    bc_type: str
        The boundary condition

    Returns
    -------
    mat: torch.Tensor with shape (*BX, nr, nr)
        The inverse of spline matrix.
    """
    nr = x.shape[-1]
    BX = x.shape[:-1]
    matshape = (*BX, nr, nr)

    device = x.device
    dtype = x.dtype

    # construct the matrix for the left hand side
    dxinv0 = 1. / (x[..., 1:] - x[..., :-1])  # (*BX,nr-1)
    zero_pad = torch.zeros_like(dxinv0[..., :1])
    dxinv = torch.cat((zero_pad, dxinv0, zero_pad), dim=-1)
    diag = (dxinv[..., :-1] + dxinv[..., 1:]) * 2  # (*BX,nr)
    offdiag = dxinv0  # (*BX,nr-1)
    spline_mat = torch.zeros(matshape, dtype=dtype, device=device)
    spdiag = spline_mat.diagonal(dim1=-2, dim2=-1)  # (*BX, nr)
    spudiag = spline_mat.diagonal(offset=1, dim1=-2, dim2=-1)
    spldiag = spline_mat.diagonal(offset=-1, dim1=-2, dim2=-1)
    spdiag[..., :] = diag
    spudiag[..., :] = offdiag
    spldiag[..., :] = offdiag

    # construct the matrix on the right hand side
    dxinv2 = (dxinv * dxinv) * 3
    diagr = (dxinv2[..., :-1] - dxinv2[..., 1:])
    udiagr = dxinv2[..., 1:-1]
    ldiagr = -udiagr
    matr = torch.zeros(matshape, dtype=dtype, device=device)
    matrdiag = matr.diagonal(dim1=-2, dim2=-1)
    matrudiag = matr.diagonal(offset=1, dim1=-2, dim2=-1)
    matrldiag = matr.diagonal(offset=-1, dim1=-2, dim2=-1)
    matrdiag[..., :] = diagr
    matrudiag[..., :] = udiagr
    matrldiag[..., :] = ldiagr

    # modify the matrices according to the boundary conditions
    if bc_type == "natural":
        pass  # set to be natural
    elif bc_type == "clamped":
        spline_mat[..., 0, :] = 0.
        spline_mat[..., 0, 0] = 1.
        spline_mat[..., -1, :] = 0.
        spline_mat[..., -1, -1] = 1.
        matr[..., 0, :] = 0.
        matr[..., -1, :] = 0.
    elif bc_type == "not-a-knot":
        dxinv00_sq = dxinv0[..., 0]**2
        dxinv01_sq = dxinv0[..., 1]**2
        dxinv0n_sq = dxinv0[..., -1]**2
        dxinv0nm1_sq = dxinv0[..., -2]**2
        dxinv00_3 = dxinv0[..., 0] * dxinv00_sq
        dxinv01_3 = dxinv0[..., 1] * dxinv01_sq
        dxinv0n_3 = dxinv0[..., -1] * dxinv0n_sq
        dxinv0nm1_3 = dxinv0[..., -2] * dxinv0nm1_sq
        spline_mat[..., 0, 0] = dxinv00_sq
        spline_mat[..., 0, 1] = dxinv00_sq - dxinv01_sq
        spline_mat[..., 0, 2] = -dxinv01_sq
        spline_mat[..., -1, -1] = -dxinv0n_sq
        spline_mat[..., -1, -2] = dxinv0nm1_sq - dxinv0n_sq
        spline_mat[..., -1, -3] = dxinv0nm1_sq
        matr[..., 0, 0] = 2 * (-dxinv00_3)
        matr[..., 0, 1] = 2 * (dxinv00_3 + dxinv01_3)
        matr[..., 0, 2] = 2 * (-dxinv01_3)
        matr[..., -1, -1] = 2 * (-dxinv0n_3)
        matr[..., -1, -2] = 2 * (dxinv0n_3 + dxinv0nm1_3)
        matr[..., -1, -3] = 2 * (-dxinv0nm1_3)
    elif bc_type == "periodic":
        dxinv01 = dxinv0[..., -1]
        dxinv00 = dxinv0[..., 0]
        spline_mat[..., 0, -2] += dxinv01
        spline_mat[..., 0, 0] += dxinv01 * 2
        spline_mat[..., -1, 1] += dxinv00
        spline_mat[..., -1, -1] += dxinv00 * 2

        dxinv01_sq3 = 3 * dxinv01 * dxinv01
        dxinv00_sq3 = 3 * dxinv00 * dxinv00
        matr[..., 0, -2] -= dxinv01_sq3
        matr[..., 0, 0] += dxinv01_sq3
        matr[..., -1, 1] += dxinv00_sq3
        matr[..., -1, -1] -= dxinv00_sq3
    else:
        raise RuntimeError("Unknown boundary condition: %s" % bc_type)

    # solve the matrix inverse
    spline_mat_inv = torch.linalg.solve(spline_mat, matr)

    # return to the shape of x
    return spline_mat_inv



import torch
import xitorch as xt
from abc import abstractmethod
from xitorch._impls.interpolate.interp_1d import _get_spline_mat_inv

"""
This file contains the cumulative sum quadrature functions.
The functions are usually used in solve_poisson method in grid objects.
"""

class BaseSQuad(xt.EditableModule):
    @abstractmethod
    def cumsum(self, y):
        """
        Cumsum the last dimension.
        """
        pass

    @abstractmethod
    def integrate(self, y):
        """
        Integrate the last dimension without keeping the dimension.
        """
        pass

    @abstractmethod
    def getparamnames(self, methodname, prefix=""):
        pass

class CubicSplineSQuad(BaseSQuad):
    r"""
    Perform integration of given sampled values by assuming it is interpolated
    with cubic spline [1]_. It is simply

    .. math::

        S = \sum_{i=0}^{N-2} \left[\frac{1}{2}(y_i+y_{i+1}) + \frac{1}{12}(y'_i - y'_{i+1})(x_{i+1}-x_i)^2\right]

    Keyword arguments
    -----------------
    bc_type: str
        Boundary condition. See :class:`xitorch.interpolate.Interp1D` with
        ``"cspline"`` method for details.

    References
    ----------
    .. [1] Mark H. Holmes, "Connections Between Cubic Splines and Quadrature Rules" (eq. 8),
           The American Mathematical Monthly, Volume 121, Issue 8, 2014.
    """

    def __init__(self, x, bc_type="natural", **unused):
        # x: (nx,)
        xshape = x.shape
        nx = xshape[-1]

        spline_mat = _get_spline_mat_inv(x, bc_type=bc_type)
        self.spline_mat = spline_mat  # (nx, nx)
        self.xshape = xshape
        self.wy = get_trapz_weights(x)  # (nx, nx)
        self.wk = get_cspline_grad_weights(x)  # (nx, nx)

    def cumsum(self, y):
        # y: (*, nx)
        # return: (*, nx)
        y1 = y.unsqueeze(-1)  # (*, nx, 1)
        ks = torch.matmul(self.spline_mat, y1)  # (*, nx, 1)
        kfactor = torch.matmul(self.wk, ks)  # (*, nx, 1)
        yfactor = torch.matmul(self.wy, y1)  # (*, nx, 1)
        res = kfactor + yfactor  # (*, nx)
        return res.squeeze(-1)

    def integrate(self, y):
        ks = torch.matmul(self.spline_mat, y.unsqueeze(-1)).squeeze(-1)  # (*, nx)
        kfactor = torch.einsum("c,...c->...", self.wk[-1], ks)
        yfactor = torch.einsum("c,...c->...", self.wy[-1], y)
        return kfactor + yfactor

    def getparamnames(self, methodname, prefix=""):
        if methodname == "cumsum" or methodname == "integrate":
            return [prefix + "spline_mat", prefix + "wk", prefix + "wy"]
        else:
            raise KeyError("%s has no %s method" % (self.__class__.__name__, methodname))

######################## weight-based cumsum quadrature ########################
class WeightBasedSQuad(BaseSQuad):
    def __init__(self, x, **options):
        # x: (nx,)
        xshape = x.shape
        nx = xshape[-1]
        x = x.reshape(-1, nx)
        self.w = self.get_weights(x, **options)  # (*, nx, nx)

    @abstractmethod
    def get_weights(self, x, **options):
        pass

    def cumsum(self, y):
        # y: (*, nx)
        # w: (*, nx, nx)
        # returns: (*, nx)
        return torch.sum(y.unsqueeze(-2) * self.w, dim=-1)

    def integrate(self, y):
        # y: (*, nx)
        # w: (*, nx, nx)
        # returns: (*,)
        return torch.sum(y * self.w[..., -1, :], dim=-1)

    def getparamnames(self, methodname, prefix=""):
        if methodname == "cumsum" or methodname == "integrate":
            return [prefix + "w"]
        else:
            raise KeyError("%s has no %s method" % (self.__class__.__name__, methodname))

class TrapzSQuad(WeightBasedSQuad):
    r"""
    Perform integration with trapezoidal rule. It is simply

    .. math::

        S = \sum_{i=0}^{N-2} \frac{1}{2}(y_i+y_{i+1})
    """

    def get_weights(self, x):
        return get_trapz_weights(x)

class SimpsonSQuad(WeightBasedSQuad):
    """
    Perform integration with composite Simpson's rule
    """

    def get_weights(self, x):
        return get_simpson_weights(x)

# @torch.jit.script
def get_trapz_weights(x: torch.Tensor) -> torch.Tensor:
    # x: (..., nx)
    # returns: (..., nx, nx)
    half_dx = (x[..., 1:] - x[..., :-1]) * 0.5  # (..., nx-1)
    nx = x.shape[-1]
    shape = list(x.shape[:-1]) + [nx, nx]
    res = torch.zeros(shape, dtype=x.dtype, device=x.device)
    for i in range(1, nx):
        res[..., i:, i - 1:i + 1] += half_dx[..., i - 1:i].unsqueeze(-1)
    return res

# @torch.jit.script
def get_simpson_weights(x: torch.Tensor) -> torch.Tensor:
    # x: (..., nx)
    # returns: (..., nx, nx)
    # ref: https://en.wikipedia.org/wiki/Simpson%27s_rule#Composite_Simpson's_rule_for_irregularly_spaced_data
    h = x[..., 1:] - x[..., :-1]  # (..., nx-1)
    h1 = h[..., 1::2]  # (..., (nx-2)//2)
    h0 = h[..., :-1:2]  # (..., (nx-2)//2)
    h1_2 = h1 * h1
    h0_2 = h0 * h0
    h1_3 = h1_2 * h1
    h0_3 = h0_2 * h0
    alpha = (2 * h1_3 - h0_3 + 3 * h0 * h1_2) / (6 * h1 * (h1 + h0))  # (..., (nx-2)//2)
    eta   = (2 * h0_3 - h1_3 + 3 * h1 * h0_2) / (6 * h0 * (h1 + h0))  # (..., (nx-2)//2)
    beta  = (h1_3 + h0_3 + 3 * h1 * h0 * (h1 + h0)) / (6 * h1 * h0)  # (..., (nx-2)//2)
    # last part (for odd parts only)
    hN1 = h[..., 2::2]  # (..., (nx-3)//2)
    hN2 = h[..., 1:-1:2]  # (..., (nx-3)//2)
    alpha_l = (2 * hN1 * hN1 + 3 * hN1 * hN2) / (6 * (hN1 + hN2))
    eta_l   = hN1 * hN1 * hN1 / (6 * hN2 * (hN1 + hN2))
    beta_l  = (hN1 * hN1 + 3 * hN1 * hN2) / (6 * hN2)

    nx = x.shape[-1]
    shape = list(x.shape[:-1]) + [nx, nx]
    res = torch.zeros(shape, dtype=x.dtype, device=x.device)
    for i in range(2, nx, 2):
        j = i // 2 - 1
        res[..., i:, i - 2] += eta[..., j:j + 1]
        res[..., i:, i - 1] += beta[..., j:j + 1]
        res[..., i:, i] += alpha[..., j:j + 1]
    for i in range(3, nx, 2):  # last part of the odd parts
        j = i // 2 - 1
        res[..., i, i - 2] += -eta_l[..., j]
        res[..., i, i - 1] += beta_l[..., j]
        res[..., i, i] += alpha_l[..., j]

    # trapezoidal rule for the part with N=1 interval
    res[..., 1, :2] = 0.5 * h[..., 0]

    return res

# @torch.jit.script
def get_cspline_grad_weights(x):
    # x: (..., nx)
    # returns: (..., nx, nx)
    dx = (x[..., 1:] - x[..., :-1])  # (..., nx-1)
    dx_factor = dx * dx / 12.  # (..., nx-1)
    sign = torch.tensor([1., -1.], dtype=x.dtype, device=x.device)
    nx = x.shape[-1]
    shape = list(x.shape[:-1]) + [nx, nx]
    res = torch.zeros(shape, dtype=x.dtype, device=x.device)
    for i in range(1, nx):
        res[..., i:, i - 1:i + 1] += dx_factor[..., i - 1:i].unsqueeze(-1) * sign
    return res



import gc
import itertools
import torch
import pytest
from typing import Mapping, List, Callable, Optional, Tuple

__all__ = ["device_dtype_float_test", "assert_no_memleak"]

def device_dtype_float_test(only64: int = False, onlycpu: bool = False,
                            additional_kwargs: Mapping[str, List] = {},
                            skip_fcn: Optional[Callable[..., Tuple[bool, str]]] = None,
                            include_complex: bool = False) -> Callable:

    # test function decorator that will setup the dtype, device, and some additional
    # kwargs values to the arguments.
    # the skip_fcn is a function that receives the same argument as the decorated
    # function and should return a tuple of bool and string.
    # the first returned value (i.e. a bool) is whether to skip that parameters
    # and the second one is the reason
    dtypes = [torch.float, torch.float64]
    devices = [torch.device("cpu"), torch.device("cuda")]
    if only64:
        dtypes = [torch.float64]
    if onlycpu or not torch.cuda.is_available():
        devices = [torch.device("cpu")]
    if include_complex:
        dtypes.extend([_get_complex_dtype(dt) for dt in dtypes])

    kwargs_vals = additional_kwargs.values()
    argnames = ",".join(["dtype", "device"] + list(additional_kwargs.keys()))
    iters = itertools.product(dtypes, devices, *kwargs_vals)

    # decide which one to skip
    if skip_fcn is None:
        params = [*iters]
    else:
        params = []
        for p in iters:
            skip, reason = skip_fcn(*p)
            param = pytest.param(*p, marks=pytest.mark.skipif(skip, reason=reason))
            params.append(param)
    return pytest.mark.parametrize(argnames, params)

# memory test functions
def assert_no_memleak(fcn: Callable, strict: bool = True, gccollect: bool = False):
    """
    Assert no memory leak when calling the function.

    Arguments
    ---------
    fcn: Callable
        A function with no input and output to be checked.
    strict: bool
        If True, then there must be no additional tensor allocated after it
        exits the function.
    gccollect: bool
        If True, then manually apply ``gc.collect()`` after the function
        execution.

    Exceptions
    ----------
    AssertionError
        Raised if there is an indication of memory leak in the function.
    """
    size0 = _get_tensor_memory()
    ntries = 10
    if strict:
        fcn()
        if gccollect:
            gc.collect()
        size = _get_tensor_memory()
        if size0 != size:
            _show_memsize(fcn, ntries, gccollect=gccollect)
        assert size0 == size
    else:
        raise NotImplementedError("Option non-strict memory leak checking has not been implemented")

def _get_complex_dtype(dtype: torch.dtype) -> torch.dtype:
    """
    Returns the complex data type that corresponds to the input data type.

    Arguments
    ---------
    dtype: torch.dtype
        Real number data type

    Returns
    -------
    torch.dtype
        The complex data type
    """
    if dtype == torch.float32:
        return torch.complex64
    elif dtype == torch.float64:
        return torch.complex128
    else:
        raise RuntimeError("Datatype %s has no complex datatype" % dtype)

def _show_memsize(fcn, ntries: int = 10, gccollect: bool = False):
    # show the memory growth
    for i in range(ntries):
        fcn()
        if gccollect:
            gc.collect()
        size = _get_tensor_memory()
        print("%3d iteration: %.7f MiB of tensors" % (i + 1, size))

def _get_tensor_memory() -> float:
    # obtain the total memory occupied by torch.Tensor in the garbage collector
    # (units in MiB)

    # obtaining all the tensor objects from the garbage collector
    tensor_objs = [obj for obj in gc.get_objects() if isinstance(obj, torch.Tensor)]

    # iterate each tensor objects uniquely and calculate the total storage
    visited_data = set([])
    total_mem = 0.0
    for tensor in tensor_objs:
        if tensor.is_sparse:
            continue

        # check if it has been visited
        storage = tensor.storage()
        data_ptr = storage.data_ptr()  # type: ignore
        if data_ptr in visited_data:
            continue
        visited_data.add(data_ptr)

        # calculate the storage occupied
        numel = storage.size()
        elmt_size = storage.element_size()
        mem = numel * elmt_size / (1024 * 1024)  # in MiB

        total_mem += mem

    return total_mem



from abc import abstractmethod
import torch
from typing import Callable, Union, Mapping, Any, Sequence
from xitorch._utils.assertfuncs import assert_fcn_params, assert_runtime
from xitorch._core.pure_function import get_pure_function, make_sibling
from xitorch._utils.misc import set_default_option, TensorNonTensorSeparator, \
    TensorPacker, get_method
from xitorch._impls.integrate.fixed_quad import leggauss
from xitorch._docstr.api_docstr import get_methods_docstr
from xitorch.debug.modes import is_debug_enabled

__all__ = ["quad"]

def quad(
        fcn: Union[Callable[..., torch.Tensor], Callable[..., Sequence[torch.Tensor]]],
        xl: Union[float, int, torch.Tensor],
        xu: Union[float, int, torch.Tensor],
        params: Sequence[Any] = [],
        bck_options: Mapping[str, Any] = {},
        method: Union[str, Callable, None] = None,
        **fwd_options) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
    r"""
    Calculate the quadrature:

    .. math::

        y = \int_{x_l}^{x_u} f(x, \theta)\ \mathrm{d}x

    Arguments
    ---------
    fcn: callable
        The function to be integrated. Its output must be a tensor with
        shape ``(*nout)`` or list of tensors.
    xl: float, int or 1-element torch.Tensor
        The lower bound of the integration.
    xu: float, int or 1-element torch.Tensor
        The upper bound of the integration.
    params: list
        Sequence of any other parameters for the function ``fcn``.
    bck_options: dict
        Options for the backward quadrature method.
    method: str or callable or None
        Quadrature method. If None, it will choose ``"leggauss"``.
    **fwd_options
        Method-specific options (see method section).

    Returns
    -------
    torch.tensor or a list of tensors
        The quadrature results with shape ``(*nout)`` or list of tensors.
    """
    # perform implementation check if debug mode is enabled
    if is_debug_enabled():
        assert_fcn_params(fcn, (xl, *params))
    if isinstance(xl, torch.Tensor):
        assert_runtime(torch.numel(xl) == 1, "xl must be a 1-element tensors")
    if isinstance(xu, torch.Tensor):
        assert_runtime(torch.numel(xu) == 1, "xu must be a 1-element tensors")
    if method is None:
        method = "leggauss"
    fwd_options["method"] = method

    out = fcn(xl, *params)
    if isinstance(out, torch.Tensor):
        dtype = out.dtype
        device = out.device
        is_tuple_out = False
    elif len(out) > 0:
        dtype = out[0].dtype
        device = out[0].device
        is_tuple_out = True
    else:
        raise RuntimeError("The output of the fcn must be non-empty")

    pfunc = get_pure_function(fcn)
    nparams = len(params)
    if is_tuple_out:
        packer = TensorPacker(out)

        @make_sibling(pfunc)
        def pfunc2(x, *params):
            y = fcn(x, *params)
            return packer.flatten(y)

        res = _Quadrature.apply(pfunc2, xl, xu, fwd_options, bck_options, nparams,
                                dtype, device, *params, *pfunc.objparams())
        return packer.pack(res)
    else:
        return _Quadrature.apply(pfunc, xl, xu, fwd_options, bck_options, nparams,
                                 dtype, device, *params, *pfunc.objparams())

class _Quadrature(torch.autograd.Function):
    # NOTE: _Quadrature method do not involve changing the state (objparams) of
    # fcn, so there is no need in using `with fcn.useobjparams(objparams)`
    # statements.
    # The function `disable_state_change()` is used to disable state change of
    # the pure function during the execution of the forward and backward
    # calculations

    @staticmethod
    def forward(ctx, fcn, xl, xu, fwd_options, bck_options, nparams,
                dtype, device, *all_params):

        with fcn.disable_state_change():

            config = fwd_options
            ctx.bck_config = set_default_option(config, bck_options)

            params = all_params[:nparams]
            objparams = all_params[nparams:]

            # convert to tensor
            xl = torch.as_tensor(xl, dtype=dtype, device=device)
            xu = torch.as_tensor(xu, dtype=dtype, device=device)

            # apply transformation if the boundaries contain inf
            if _isinf(xl) or _isinf(xu):
                tfm = _TanInfTransform()

                @make_sibling(fcn)
                def fcn2(t, *params):
                    ys = fcn(tfm.forward(t), *params)
                    dxdt = tfm.dxdt(t)
                    return ys * dxdt

                tl = tfm.x2t(xl)
                tu = tfm.x2t(xu)
            else:
                fcn2 = fcn
                tl = xl
                tu = xu

            method = config.pop("method")
            methods = {
                "leggauss": leggauss
            }
            method_fcn = get_method("quad", methods, method)
            y = method_fcn(fcn2, tl, tu, params, **config)

            # save the parameters for backward
            ctx.param_sep = TensorNonTensorSeparator(all_params)
            tensor_params = ctx.param_sep.get_tensor_params()
            ctx.xltensor = isinstance(xl, torch.Tensor)
            ctx.xutensor = isinstance(xu, torch.Tensor)
            xlxu_tensor = ([xl] if ctx.xltensor else []) + \
                          ([xu] if ctx.xutensor else [])
            ctx.xlxu_nontensor = ([xl] if not ctx.xltensor else []) + \
                                 ([xu] if not ctx.xutensor else [])
            ctx.save_for_backward(*xlxu_tensor, *tensor_params)
            ctx.fcn = fcn
            ctx.nparams = nparams
            return y

    @staticmethod
    def backward(ctx, grad_ys):
        # retrieve the params
        ntensor_params = ctx.param_sep.ntensors()
        tensor_params = ctx.saved_tensors[-ntensor_params:]
        allparams = ctx.param_sep.reconstruct_params(tensor_params)
        nparams = ctx.nparams
        params = allparams[:nparams]
        fcn = ctx.fcn

        with fcn.disable_state_change():

            # restore xl, and xu
            xlxu_tensor = ctx.saved_tensors[:-ntensor_params]
            if ctx.xltensor and ctx.xutensor:
                xl, xu = xlxu_tensor
            elif ctx.xltensor:
                xl = xlxu_tensor[0]
                xu = ctx.xlxu_nontensor[0]
            elif ctx.xutensor:
                xu = xlxu_tensor[0]
                xl = ctx.xlxu_nontensor[0]
            else:
                xl, xu = ctx.xlxu_nontensor

            # calculate the gradient for the boundaries
            grad_xl = -torch.dot(grad_ys.reshape(-1), fcn(xl, *params).reshape(-1)
                                 ).reshape(xl.shape) if ctx.xltensor else None
            grad_xu = torch.dot(grad_ys.reshape(-1), fcn(xu, *params).reshape(-1)
                                ).reshape(xu.shape) if ctx.xutensor else None

            def new_fcn(x, *grad_y_params):
                grad_ys = grad_y_params[0]
                # not setting objparams and params because the params and objparams
                # are still the same objects as the objects outside
                with torch.enable_grad():
                    f = fcn(x, *params)
                dfdts = torch.autograd.grad(f, tensor_params,
                                            grad_outputs=grad_ys,
                                            retain_graph=True,
                                            create_graph=torch.is_grad_enabled())
                return dfdts

            # reconstruct grad_params
            # listing tensor_params in the params of quad to make sure it gets
            # the gradient calculated
            dydts = quad(new_fcn, xl, xu, params=(grad_ys, *tensor_params),
                         fwd_options=ctx.bck_config, bck_options=ctx.bck_config)
            dydns = [None for _ in range(ctx.param_sep.nnontensors())]
            grad_params = ctx.param_sep.reconstruct_params(dydts, dydns)

            return (None, grad_xl, grad_xu, None, None, None, None, None, *grad_params)

def _isinf(x):
    return torch.any(torch.isinf(x))

class _BaseInfTransform(object):
    @abstractmethod
    def forward(self, t):
        pass

    @abstractmethod
    def dxdt(self, t):
        pass

    @abstractmethod
    def x2t(self, x):
        pass

class _TanInfTransform(_BaseInfTransform):
    def forward(self, t):
        return torch.tan(t)

    def dxdt(self, t):
        sec = 1. / torch.cos(t)
        return sec * sec

    def x2t(self, x):
        return torch.atan(x)


# docstring completion
quad.__doc__ = get_methods_docstr(quad, [leggauss])



import torch
import warnings
from typing import Callable, List, Any, Union, Sequence
from xitorch import LinearOperator
from xitorch._core.pure_function import get_pure_function, make_sibling, PureFunction
from xitorch._utils.assertfuncs import assert_type
from xitorch._utils.misc import TensorNonTensorSeparator

__all__ = ["jac", "hess"]

def jac(fcn: Callable[..., torch.Tensor], params: Sequence[Any],
        idxs: Union[None, int, Sequence[int]] = None) -> Union[LinearOperator, List]:
    """
    Returns the LinearOperator that acts as the jacobian of the params.
    The shape of LinearOperator is (nout, nin) where `nout` and `nin` are the
    total number of elements in the output and the input, respectively.

    Arguments
    ---------
    * fcn: Callable[...,torch.Tensor]
        Callable with tensor output and arbitrary numbers of input parameters.
    * params: Sequence[Any]
        List of input parameters of the function.
    * idxs: int or list of int or None
        List of the parameters indices to get the jacobian.
        The pointed parameters in `params` must be tensors and requires_grad.
        If it is None, then it will return all jacobian for all parameters that
        are tensor which requires_grad.

    Returns
    -------
    * linops: list of LinearOperator
        List of LinearOperator of the jacobian
    """
    # check idxs
    idxs_list = _setup_idxs(idxs, params)

    # make the function a functional (depends on all parameters in the object)
    pfcn = get_pure_function(fcn)
    res = [_Jac(pfcn, params, idx) for idx in idxs_list]
    if isinstance(idxs, int):
        return res[0]
    return res

def hess(fcn: Callable[..., torch.Tensor], params: Sequence[Any],
         idxs: Union[None, int, Sequence[int]] = None) -> Union[LinearOperator, List]:
    """
    Returns the LinearOperator that acts as the Hessian of the params.
    The shape of LinearOperator is (nin, nin) where `nin` is the
    total number of elements in the input.

    Arguments
    ---------
    * fcn: Callable[...,torch.Tensor]
        Callable with tensor output and arbitrary numbers of input parameters.
        The numel of the output must be 1.
    * params: Sequence[Any]
        List of input parameters of the function.
    * idxs: int or list of int or None
        List of the parameters indices to get the jacobian.
        The pointed parameters in `params` must be tensors and requires_grad.
        If it is None, then it will return all Hessian for all parameters that
        are tensor which requires_grad.

    Returns
    -------
    * linops: list of LinearOperator
        List of LinearOperator of the Hessian
    """
    idxs_list = _setup_idxs(idxs, params)

    # make the function a functional (depends on all parameters in the object)
    pfcn = get_pure_function(fcn)

    res = []

    def gen_pfcn2(idx):
        @make_sibling(pfcn)
        def pfcn2(*params):
            with torch.enable_grad():
                z = pfcn(*params)
            grady, = torch.autograd.grad(z, (params[idx],), retain_graph=True,
                                         create_graph=torch.is_grad_enabled())
            return grady
        return pfcn2

    for idx in idxs_list:
        # suppress warnings of double implementation in hermitian matrix
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            hs = _Jac(gen_pfcn2(idx), params, idx, is_hermitian=True)
        res.append(hs)

    if isinstance(idxs, int):
        return res[0]
    return res

class _Jac(LinearOperator):
    def __init__(self, fcn: PureFunction,
                 params: Sequence[Any], idx: int, is_hermitian=False) -> None:

        # TODO: check if fcn has kwargs

        # run once to get the shapes and numels
        yparam = params[idx]
        with torch.enable_grad():
            yout = fcn(*params)  # (*nout)
            v = torch.ones_like(yout).to(yout.device).requires_grad_()  # (*nout)
            dfdy, = torch.autograd.grad(yout, (yparam,), grad_outputs=v, create_graph=True)  # (*nin)

        inshape = yparam.shape
        outshape = yout.shape
        nin = torch.numel(yparam)
        nout = torch.numel(yout)

        super(_Jac, self).__init__(
            shape=(nout, nin),
            is_hermitian=is_hermitian,
            dtype=yparam.dtype,
            device=yparam.device)

        self.fcn = fcn
        self.yparam = yparam
        self.params = list(params)
        self.objparams = fcn.objparams()
        self.yout = yout
        self.v = v
        self.idx = idx
        self.dfdy = dfdy
        self.inshape = inshape
        self.outshape = outshape
        self.nin = nin
        self.nout = nout

        # params tensor is the LinearOperator's parameters
        self.param_sep = TensorNonTensorSeparator(params)
        self.params_tensor = self.param_sep.get_tensor_params()
        self.id_params_tensor = [id(param) for param in self.params_tensor]
        self.id_objparams_tensor = [id(param) for param in self.objparams]

    def _getparamnames(self, prefix: str = "") -> List[str]:
        return [prefix + "yparam"] + \
               [prefix + ("params_tensor[%d]" % i) for i in range(len(self.params_tensor))] + \
               [prefix + ("objparams[%d]" % i) for i in range(len(self.objparams))]

    def _mv(self, gy: torch.Tensor) -> torch.Tensor:
        # gy: (..., nin)
        # returns: (..., nout)

        # if the object parameter is still the same, then use the pre-calculated values
        if self.__param_tensors_unchanged():
            v = self.v
            dfdy = self.dfdy
        # otherwise, reevaluate by replacing the parameters with the new tensor params
        else:
            with torch.enable_grad(), self.fcn.useobjparams(self.objparams):
                self.__update_params()
                yparam = self.params[self.idx]
                yout = self.fcn(*self.params)  # (*nout)
                v = torch.ones_like(yout).to(yout.device).requires_grad_()  # (*nout)
                dfdy, = torch.autograd.grad(yout, (yparam,), grad_outputs=v, create_graph=True)  # (*nin)

        gy1 = gy.reshape(-1, self.nin)  # (nbatch, nin)
        nbatch = gy1.shape[0]
        dfdyfs_list = []
        for i in range(nbatch):
            dfdyf, = torch.autograd.grad(dfdy, (v,), grad_outputs=gy1[i].reshape(self.inshape),
                                         retain_graph=True, create_graph=torch.is_grad_enabled())  # (*nout)
            dfdyfs_list.append(dfdyf.unsqueeze(0))
        dfdyfs = torch.cat(dfdyfs_list, dim=0)  # (nbatch, *nout)

        res = dfdyfs.reshape(*gy.shape[:-1], self.nout)  # (..., nout)
        res = connect_graph(res, self.params_tensor)
        res = connect_graph(res, self.objparams)
        return res

    def _rmv(self, gout: torch.Tensor) -> torch.Tensor:
        # gout: (..., nout)
        # self.yfcn: (*nin)
        if self.__param_tensors_unchanged():
            yout = self.yout
            yparam = self.yparam
        else:
            with torch.enable_grad(), self.fcn.useobjparams(self.objparams):
                self.__update_params()
                yparam = self.params[self.idx]
                yout = self.fcn(*self.params)  # (*nout)

        gout1 = gout.reshape(-1, self.nout)  # (nbatch, nout)
        nbatch = gout1.shape[0]
        dfdy_list = []
        for i in range(nbatch):
            one_dfdy, = torch.autograd.grad(yout, (yparam,), grad_outputs=gout1[i].reshape(self.outshape),
                                            retain_graph=True, create_graph=torch.is_grad_enabled())  # (*nin)
            dfdy_list.append(one_dfdy.unsqueeze(0))
        dfdy = torch.cat(dfdy_list, dim=0)  # (nbatch, *nin)

        res = dfdy.reshape(*gout.shape[:-1], self.nin)  # (..., nin)
        res = connect_graph(res, self.params_tensor)
        res = connect_graph(res, self.objparams)
        return res  # (..., nin)

    def __param_tensors_unchanged(self):
        return [id(param) for param in self.params_tensor] == self.id_params_tensor and \
               [id(param) for param in self.objparams] == self.id_objparams_tensor

    def __update_params(self):
        self.params = self.param_sep.reconstruct_params(self.params_tensor)

def connect_graph(out, params):
    # just to have a dummy graph, in case there is a parameter that
    # is disconnected in calculating df/dy
    return out + sum([p.reshape(-1)[0] * 0 for p in params])

def _setup_idxs(idxs, params):
    if idxs is None:
        idxs = [i for i, t in enumerate(params) if isinstance(t, torch.Tensor) and t.requires_grad]
    elif isinstance(idxs, int):
        idxs = [idxs]

    for p in idxs:
        assert_type(isinstance(params[p], torch.Tensor) and params[p].requires_grad,
                    "The %d-th element (0-based) must be a tensor which requires grad" % p)
    return idxs
